#
# (C) Copyright IBM Corp. 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import pkgutil


def main(args):
    print("Extracting preinstalled Python modules...")
    runtime_meta = dict()
    mods = list(pkgutil.iter_modules())
    runtime_meta['preinstalls'] = [entry for entry in sorted([[mod, is_pkg] for _, mod, is_pkg in mods])]
    python_version = sys.version_info
    runtime_meta['python_ver'] = str(python_version[0])+"."+str(python_version[1])

    return runtime_meta



import os
import time
import pickle
import logging
import inspect
import pywren_ibm_cloud as pywren
from pywren_ibm_cloud import utils
from pywren_ibm_cloud.wait import wait
from pywren_ibm_cloud.runtime import select_runtime
from pywren_ibm_cloud.job.serialize import SerializeIndependent, create_module_data
from pywren_ibm_cloud.partitioner import create_partitions, partition_processor
from pywren_ibm_cloud.storage.storage_utils import create_func_key, create_agg_data_key
from pywren_ibm_cloud.wrenconfig import RUNTIME_TIMEOUT, RUNTIME_RI_MEMORY_DEFAULT, MAX_AGG_DATA_SIZE

logger = logging.getLogger(__name__)


def create_call_async_job(config, internal_storage, executor_id, func, data, extra_env=None,
                          extra_meta=None, runtime_memory=None, runtime_timeout=RUNTIME_TIMEOUT):
    """
    Wrapper to create call_async job that contains only one function invocation.
    """
    return _create_job(config, internal_storage, executor_id, func, [data], extra_env=extra_env,
                       extra_meta=extra_meta, runtime_memory=runtime_memory, runtime_timeout=runtime_timeout)


def create_map_job(config, internal_storage, executor_id, map_function, iterdata, obj_chunk_size=None,
                   extra_env=None, extra_meta=None, runtime_name=None, runtime_memory=None, remote_invocation=False,
                   remote_invocation_groups=None, invoke_pool_threads=128, exclude_modules=None, is_cf_cluster=False,
                   runtime_timeout=RUNTIME_TIMEOUT, overwrite_invoke_args=None):
    """
    Wrapper to create a map job.  It integrates COS logic to process objects.
    """
    data = utils.iterdata_as_list(iterdata)
    map_func = map_function
    map_iterdata = data
    new_invoke_pool_threads = invoke_pool_threads
    new_runtime_memory = runtime_memory

    # Object processing functionality
    parts_per_object = None
    if utils.is_object_processing_function(map_function):
        '''
        If it is object processing function, create partitions according chunk_size
        '''
        logger.debug("Calling map on partitions from object storage flow")
        arg_data = utils.verify_args(map_function, data, object_processing=True)
        map_iterdata, parts_per_object = create_partitions(config, arg_data, obj_chunk_size)
        map_func = partition_processor(map_function)
    # ########

    # Remote invocation functionality
    original_iterdata_len = len(map_iterdata)
    if original_iterdata_len == 1 or is_cf_cluster:
        remote_invocation = False
    if remote_invocation:
        rabbitmq_monitor = "PYWREN_RABBITMQ_MONITOR" in os.environ

        def remote_invoker(input_data):
            pw = pywren.ibm_cf_executor(runtime=runtime_name,
                                        rabbitmq_monitor=rabbitmq_monitor)
            return pw.map(map_function, input_data,
                          runtime_memory=runtime_memory,
                          invoke_pool_threads=invoke_pool_threads,
                          extra_env=extra_env,
                          extra_meta=extra_meta)

        map_func = remote_invoker
        if remote_invocation_groups:
            map_iterdata = [[iterdata[x:x+remote_invocation_groups]]
                            for x in range(0, original_iterdata_len, remote_invocation_groups)]
        else:
            map_iterdata = [iterdata]
        new_invoke_pool_threads = 1
        new_runtime_memory = RUNTIME_RI_MEMORY_DEFAULT
    # ########

    job_description = _create_job(config, internal_storage, executor_id,
                                  map_func, map_iterdata,
                                  extra_env=extra_env,
                                  extra_meta=extra_meta,
                                  runtime_memory=new_runtime_memory,
                                  invoke_pool_threads=new_invoke_pool_threads,
                                  overwrite_invoke_args=overwrite_invoke_args,
                                  exclude_modules=exclude_modules,
                                  original_func_name=map_function.__name__,
                                  remote_invocation=remote_invocation,
                                  original_iterdata_len=original_iterdata_len,
                                  runtime_timeout=runtime_timeout)

    return job_description, parts_per_object


def create_reduce_job(config, internal_storage, executor_id, reduce_function, reduce_runtime_memory,
                      map_futures, parts_per_object, reducer_one_per_object, extra_env, extra_meta):
    """
    Wrapper to create a reduce job. Apply a function across all map futures.
    """
    map_iterdata = [[map_futures, ]]

    if parts_per_object and reducer_one_per_object:
        prev_total_partitons = 0
        map_iterdata = []
        for total_partitions in parts_per_object:
            map_iterdata.append([map_futures[prev_total_partitons:prev_total_partitons+total_partitions]])
            prev_total_partitons = prev_total_partitons + total_partitions

    def reduce_function_wrapper(fut_list, internal_storage, ibm_cos):
        logger.info('Waiting for results')
        if 'SHOW_MEMORY_USAGE' in os.environ:
            show_memory = eval(os.environ['SHOW_MEMORY_USAGE'])
        else:
            show_memory = False
        # Wait for all results
        wait(fut_list, executor_id, internal_storage, download_results=True)
        results = [f.result() for f in fut_list if f.done and not f.futures]
        fut_list.clear()
        reduce_func_args = {'results': results}

        if show_memory:
            logger.debug("Memory usage after getting the results: {}".format(utils.get_current_memory_usage()))

        # Run reduce function
        func_sig = inspect.signature(reduce_function)
        if 'ibm_cos' in func_sig.parameters:
            reduce_func_args['ibm_cos'] = ibm_cos

        return reduce_function(**reduce_func_args)

    return _create_job(config, internal_storage, executor_id, reduce_function_wrapper, map_iterdata,
                       runtime_memory=reduce_runtime_memory, extra_env=extra_env, extra_meta=extra_meta,
                       original_func_name=reduce_function.__name__)


def _agg_data(data_strs):
    """
    Auxiliary function that aggregates data of a job to a single byte string
    """
    ranges = []
    pos = 0
    for datum in data_strs:
        datum_len = len(datum)
        ranges.append((pos, pos+datum_len-1))
        pos += datum_len
    return b"".join(data_strs), ranges


def _create_job(config, internal_storage, executor_id, func, iterdata, extra_env=None, extra_meta=None,
                runtime_name=None, runtime_memory=None, invoke_pool_threads=128, overwrite_invoke_args=None,
                exclude_modules=None, original_func_name=None, remote_invocation=False, original_iterdata_len=None,
                runtime_timeout=RUNTIME_TIMEOUT):
    """
    :param func: the function to map over the data
    :param iterdata: An iterable of input data
    :param extra_env: Additional environment variables for CF environment. Default None.
    :param extra_meta: Additional metadata to pass to CF. Default None.
    :param remote_invocation: Enable remote invocation. Default False.
    :param invoke_pool_threads: Number of threads to use to invoke.
    :param data_all_as_one: upload the data as a single object. Default True
    :param overwrite_invoke_args: Overwrite other args. Mainly used for testing.
    :param exclude_modules: Explicitly keep these modules from pickled dependencies.
    :param original_func_name: Name of the function to invoke.
    :return: A list with size `len(iterdata)` of futures for each job
    :rtype:  list of futures.
    """
    log_level = os.getenv('PYWREN_LOG_LEVEL')

    if runtime_name is None:
        runtime_name = config['pywren']['runtime']
    if runtime_memory is None:
        runtime_memory = config['pywren']['runtime_memory']
    runtime_memory = int(runtime_memory)

    runtime_preinstalls = select_runtime(config, internal_storage, executor_id,
                                         runtime_name, runtime_memory)
    serializer = SerializeIndependent(runtime_preinstalls)

    if original_func_name:
        func_name = original_func_name
    else:
        func_name = func.__name__

    data = utils.iterdata_as_list(iterdata)

    if extra_env is not None:
        extra_env = utils.convert_bools_to_string(extra_env)

    if not data:
        return []

    # This allows multiple parameters in functions
    data = utils.verify_args(func, data)

    callgroup_id = utils.create_callgroup_id()

    host_job_meta = {}
    job_description = {}

    job_description['runtime_name'] = runtime_name
    job_description['runtime_memory'] = runtime_memory
    job_description['runtime_timeout'] = runtime_timeout
    job_description['func_name'] = func_name
    job_description['extra_env'] = extra_env
    job_description['extra_meta'] = extra_meta
    job_description['total_data_items'] = len(data)
    job_description['invoke_pool_threads'] = invoke_pool_threads
    job_description['overwrite_invoke_args'] = overwrite_invoke_args
    job_description['callgroup_id'] = callgroup_id
    job_description['remote_invocation'] = remote_invocation
    job_description['original_iterdata_len'] = original_iterdata_len

    log_msg = 'ExecutorID {} Serializing function and data'.format(executor_id)
    logger.debug(log_msg)
    # pickle func and all data (to capture module dependencies)
    func_and_data_ser, mod_paths = serializer([func] + data)

    func_str = func_and_data_ser[0]
    data_strs = func_and_data_ser[1:]
    data_size_bytes = sum(len(x) for x in data_strs)

    host_job_meta['agg_data'] = False
    host_job_meta['data_size_bytes'] = data_size_bytes

    log_msg = 'ExecutorID {} - Uploading function and data'.format(executor_id)
    logger.info(log_msg)
    if not log_level:
        print(log_msg, end=' ')

    if data_size_bytes < MAX_AGG_DATA_SIZE:
        agg_data_key = create_agg_data_key(internal_storage.prefix, executor_id, callgroup_id)
        job_description['data_key'] = agg_data_key
        agg_data_bytes, agg_data_ranges = _agg_data(data_strs)
        job_description['data_ranges'] = agg_data_ranges
        agg_upload_time = time.time()
        internal_storage.put_data(agg_data_key, agg_data_bytes)
        host_job_meta['agg_data'] = True
        host_job_meta['data_upload_time'] = time.time() - agg_upload_time
        host_job_meta['data_upload_timestamp'] = time.time()
    else:
        log_msg = ('ExecutorID {} - Total data exceeded '
                   'maximum size of {} bytes'.format(executor_id, MAX_AGG_DATA_SIZE))
        raise Exception(log_msg)

    if exclude_modules:
        for module in exclude_modules:
            for mod_path in list(mod_paths):
                if module in mod_path and mod_path in mod_paths:
                    mod_paths.remove(mod_path)

    module_data = create_module_data(mod_paths)
    # Create func and upload
    host_job_meta['func_name'] = func_name
    func_module_str = pickle.dumps({'func': func_str, 'module_data': module_data}, -1)
    host_job_meta['func_module_bytes'] = len(func_module_str)

    func_upload_time = time.time()
    func_key = create_func_key(internal_storage.prefix, executor_id, callgroup_id)
    job_description['func_key'] = func_key
    internal_storage.put_func(func_key, func_module_str)
    host_job_meta['func_upload_time'] = time.time() - func_upload_time
    host_job_meta['func_upload_timestamp'] = time.time()

    if not log_level:
        func_and_data_size = utils.sizeof_fmt(host_job_meta['func_module_bytes']+host_job_meta['data_size_bytes'])
        log_msg = '- Total: {}'.format(func_and_data_size)
        print(log_msg)

    job_description['host_job_meta'] = host_job_meta

    return job_description



import logging.config


def default_logging_config(log_level='INFO'):
    if log_level == 'DEBUG_BOTO3':
        log_level = 'DEBUG'
        logging.getLogger('ibm_boto3').setLevel(logging.DEBUG)
        logging.getLogger('ibm_botocore').setLevel(logging.DEBUG)

    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'standard'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': log_level,
                'propagate': True
            }
        }
    })


def ibm_cf_logging_config(log_level='INFO'):
    if log_level == 'DEBUG_BOTO3':
        log_level = 'DEBUG'
        logging.getLogger('ibm_boto3').setLevel(logging.DEBUG)
        logging.getLogger('ibm_botocore').setLevel(logging.DEBUG)

    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '[%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
                'stream': 'ext://sys.stdout'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': log_level,
                'propagate': True
            }
        }
    })



import os
import time
import random
import logging
import threading
from .backends.ibm_cf import IbmCfComputeBackend

logger = logging.getLogger(__name__)


class ThreadSafeSingleton(type):
    _instances = {}
    _singleton_lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._singleton_lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(ThreadSafeSingleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Compute(metaclass=ThreadSafeSingleton):
    """
    An InternalCompute object is used by invokers and other components to access underlying compute backend
    without exposing the the implementation details.
    """

    def __init__(self, compute_config):
        self.log_level = os.getenv('PYWREN_LOG_LEVEL')
        self.compute_config = compute_config

        self.compute_backend = compute_config['compute_backend']

        if self.compute_backend == 'ibm_cf':
            self.compute_handler = IbmCfComputeBackend(compute_config['ibm_cf'])

        else:
            raise NotImplementedError(("Using {} as compute backend is" +
                                       "not supported yet").format(self.compute_backend))

    def invoke(self, runtime_name, runtime_memory, payload):
        """
        Invoke -- return information about this invocation
        """
        act_id = self.compute_handler.invoke(runtime_name, runtime_memory, payload)
        attempts = 1

        while not act_id and self.invocation_retry and attempts < self.retries:
            attempts += 1
            selected_sleep = random.choice(self.retry_sleeps)
            exec_id = payload['executor_id']
            call_id = payload['call_id']
            log_msg = ('ExecutorID {} - Function {} - Retry {} in {} seconds'.format(exec_id, call_id, attempts, selected_sleep))
            logger.debug(log_msg)
            time.sleep(selected_sleep)
            act_id = self.compute_handler.invoke(runtime_name, runtime_memory, payload)

        return act_id

    def invoke_with_result(self, runtime_name, runtime_memory, payload={}):
        """
        Invoke waiting for a result -- return information about this invocation
        """
        return self.compute_handler.invoke_with_result(runtime_name, runtime_memory, payload)

    def create_runtime(self, docker_image_name, memory, code=None, is_binary=True, timeout=300000):
        """
        Wrapper method to create a runtime in the compute backend.
        return: the name of the runtime
        """
        return self.compute_handler.create_runtime(docker_image_name, memory, code=code,
                                                   is_binary=is_binary, timeout=timeout)

    def delete_runtime(self, docker_image_name, memory):
        """
        Wrapper method to create a runtime in the compute backend
        """
        self.compute_handler.delete_runtime(docker_image_name, memory)

    def delete_all_runtimes(self):
        """
        Wrapper method to create a runtime in the compute backend
        """
        self.compute_handler.delete_all_runtimes()

    def list_runtimes(self, docker_image_name='all'):
        """
        Wrapper method to list deployed runtime in the compute backend
        """
        return self.compute_handler.list_runtimes(docker_image_name=docker_image_name)

    def get_runtime_key(self, docker_image_name, memory):
        """
        Wrapper method that returns a formated string that represents the runtime key.
        Each backend has its own runtime key format. Used to store modules preinstalls into storage
        """
        return self.compute_handler.get_runtime_key(docker_image_name, memory)



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class StorageNoSuchKeyError(Exception):
    def __init__(self, key):
        msg = "No such key {} found in storage.".format(key)
        super(StorageNoSuchKeyError, self).__init__(msg)


class StorageOutputNotFoundError(Exception):
    def __init__(self, executor_id, call_id):
        msg = "Output for {} {} not found in storage.".format(executor_id, call_id)
        super(StorageOutputNotFoundError, self).__init__(msg)


class StorageConfigMismatchError(Exception):
    def __init__(self, current_path, prev_path):
        msg = "The data is stored at {}, but current storage is configured at {}.".format(
            prev_path, current_path)
        super(StorageConfigMismatchError, self).__init__(msg)



#
# (C) Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import shutil
import logging
import zipfile
from pywren_ibm_cloud import wrenconfig
from pywren_ibm_cloud.storage import InternalStorage
from pywren_ibm_cloud.compute import Compute
from pywren_ibm_cloud.utils import version_str

logger = logging.getLogger(__name__)

ZIP_LOCATION = os.path.join(os.getcwd(), 'ibmcf_pywren.zip')


def _get_default_image_name():
    this_version_str = version_str(sys.version_info)
    if this_version_str == '3.5':
        image_name = wrenconfig.RUNTIME_DEFAULT_35
    elif this_version_str == '3.6':
        image_name = wrenconfig.RUNTIME_DEFAULT_36
    elif this_version_str == '3.7':
        image_name = wrenconfig.RUNTIME_DEFAULT_37
    return image_name


def _get_pywren_location():
    my_location = os.path.dirname(os.path.abspath(__file__))
    pw_location = os.path.join(my_location, '..')
    return pw_location


def _create_zip_action():
    logger.debug("Creating zip action in {}".format(ZIP_LOCATION))

    def add_folder_to_zip(zip_file, full_dir_path, sub_dir=''):
        for file in os.listdir(full_dir_path):
            full_path = os.path.join(full_dir_path, file)
            if os.path.isfile(full_path):
                zip_file.write(full_path, os.path.join('pywren_ibm_cloud', sub_dir, file), zipfile.ZIP_DEFLATED)
            elif os.path.isdir(full_path) and '__pycache__' not in full_path:
                add_folder_to_zip(zip_file, full_path, os.path.join(sub_dir, file))

    try:
        pywren_location = _get_pywren_location()

        with zipfile.ZipFile(ZIP_LOCATION, 'w') as ibmcf_pywren_zip:
            main_file = os.path.join(pywren_location, 'action', '__main__.py')
            ibmcf_pywren_zip.write(main_file, '__main__.py', zipfile.ZIP_DEFLATED)
            add_folder_to_zip(ibmcf_pywren_zip, pywren_location)
    except Exception:
        raise Exception('Unable to create the {} action package'.format(ZIP_LOCATION))


def _extract_modules(docker_image_name, internal_compute):
    # Extract installed Python modules from docker image
    pywren_location = _get_pywren_location()
    action_location = os.path.join(pywren_location, "runtime", "extract_preinstalls_fn.py")

    with open(action_location, "r") as action_py:
        action_code = action_py.read()

    memory = 192

    # old_stdout = sys.stdout
    # sys.stdout = open(os.devnull, 'w')
    internal_compute.create_runtime(docker_image_name, memory, code=action_code, is_binary=False)
    # sys.stdout = old_stdout
    logger.debug("Extracting Python modules list from: {}".format(docker_image_name))
    try:
        runtime_meta = internal_compute.invoke_with_result(docker_image_name, memory)
    except Exception:
        raise("Unable to invoke 'modules' action")
    try:
        internal_compute.delete_runtime(docker_image_name, memory)
    except Exception:
        raise("Unable to delete 'modules' action")

    if 'preinstalls' not in runtime_meta:
        raise Exception(runtime_meta)

    return runtime_meta


def _create_blackbox_runtime(docker_image_name, memory, timeout, runtime_meta, internal_compute, internal_storage):
    # Create runtime_name from runtime_name docker image
    # Upload zipped PyWren action
    with open(ZIP_LOCATION, "rb") as action_zip:
        action_bin = action_zip.read()

    internal_compute.create_runtime(docker_image_name, memory, code=action_bin, timeout=timeout)

    if runtime_meta is not None:
        try:
            runtime_key = internal_compute.get_runtime_key(docker_image_name, memory)
            internal_storage.put_runtime_info(runtime_key, runtime_meta)
        except Exception:
            raise("Unable to upload 'preinstalled modules' file into {}".format(internal_storage.storage_backend))


def create_runtime(docker_image_name, memory=None, config=None):
    if docker_image_name == 'default':
        docker_image_name = _get_default_image_name()
    logger.info('Creating new PyWren runtime based on Docker image {}'.format(docker_image_name))

    config = wrenconfig.default(config)
    storage_config = wrenconfig.extract_storage_config(config)
    internal_storage = InternalStorage(storage_config)
    compute_config = wrenconfig.extract_compute_config(config)
    internal_compute = Compute(compute_config)

    _create_zip_action()

    runtime_meta = _extract_modules(docker_image_name, internal_compute)
    memory = config['pywren']['runtime_memory'] if not memory else memory
    timeout = config['pywren']['runtime_timeout']
    _create_blackbox_runtime(docker_image_name, memory, timeout, runtime_meta,
                             internal_compute, internal_storage)


def update_runtime(docker_image_name, config=None):
    if docker_image_name == 'default':
        docker_image_name = _get_default_image_name()

    logger.info('Updating runtime: {}'.format(docker_image_name))

    config = wrenconfig.default(config)
    storage_config = wrenconfig.extract_storage_config(config)
    internal_storage = InternalStorage(storage_config)
    compute_config = wrenconfig.extract_compute_config(config)
    internal_compute = Compute(compute_config)
    _create_zip_action()

    timeout = config['pywren']['runtime_timeout']

    if docker_image_name != 'all':
        runtime_meta = _extract_modules(docker_image_name, internal_compute)
    else:
        runtime_meta = None

    runtimes = internal_compute.list_runtimes(docker_image_name)

    for runtime in runtimes:
        _create_blackbox_runtime(runtime[0], runtime[1], timeout, runtime_meta,
                                 internal_compute, internal_storage)


def build_runtime(docker_image_name, config=None):
    logger.info('Creating a new docker image from Dockerfile')
    logger.info('Docker image name: {}'.format(docker_image_name))

    cmd = 'docker build -t {} .'.format(docker_image_name)
    res = os.system(cmd)
    if res != 0:
        exit()

    cmd = 'docker push {}'.format(docker_image_name)
    res = os.system(cmd)
    if res != 0:
        exit()

    create_runtime(docker_image_name, config=config)
    update_runtime(docker_image_name, config=config)


def delete_runtime(docker_image_name, config=None):
    if docker_image_name == 'default':
        docker_image_name = _get_default_image_name()
    logger.info('Deleting runtimes based on Docker image: {}'.format(docker_image_name))

    config = wrenconfig.default(config)
    storage_config = wrenconfig.extract_storage_config(config)
    internal_storage = InternalStorage(storage_config)
    compute_config = wrenconfig.extract_compute_config(config)
    internal_compute = Compute(compute_config)

    runtimes = internal_compute.list_runtimes(docker_image_name)
    for runtime in runtimes:
        internal_compute.delete_runtime(runtime[0], runtime[1])
        runtime_key = internal_compute.get_runtime_key(runtime[0], runtime[1])
        internal_storage.delete_runtime_info(runtime_key)


def clean_runtimes(config=None):
    logger.info('Cleaning all runtimes')
    config = wrenconfig.default(config)
    storage_config = wrenconfig.extract_storage_config(config)
    internal_storage = InternalStorage(storage_config)
    compute_config = wrenconfig.extract_compute_config(config)
    internal_compute = Compute(compute_config)

    # Clean local runtime_meta cache
    LOCAL_HOME_DIR = os.path.expanduser('~')
    cache_dir = os.path.join(LOCAL_HOME_DIR, '.pywren')
    if os.path.exists(cache_dir):
        shutil.rmtree(cache_dir)

    sh = internal_storage.storage_handler
    runtimes = sh.list_keys_with_prefix(storage_config['storage_bucket'], 'runtime')
    if runtimes:
        sh.delete_objects(storage_config['storage_bucket'], runtimes)

    internal_compute.delete_all_runtimes()



import sys
import unittest
import pywren_ibm_cloud as pywren
import urllib.request
from pywren_ibm_cloud.storage import InternalStorage
from pywren_ibm_cloud import wrenconfig
from multiprocessing.pool import ThreadPool
import logging


# logging.basicConfig(level=logging.DEBUG)


def initTests():
    print('Uploading test files...')

    def up(param):
        i, url = param
        content = urllib.request.urlopen(url).read()
        STORAGE.put_object(bucket_name=STORAGE_CONFIG['storage_bucket'],
                           key=f'{PREFIX}/test{str(i)}',
                           data=content)
        return len(content.split())

    pool = ThreadPool(128)
    results = pool.map(up, enumerate(TEST_FILES_URLS))
    pool.close()
    pool.join()
    result_to_compare = 1 + sum(results)  # including result's word

    STORAGE.put_object(bucket_name=STORAGE_CONFIG['storage_bucket'],
                       key=f'{PREFIX}/result',
                       data=str(result_to_compare).encode())


def list_test_keys():
    return STORAGE.list_keys_with_prefix(bucket_name=STORAGE_CONFIG['storage_bucket'],
                                         prefix=PREFIX)


def cleanTests():
    print('Deleting test files...')
    for key in list_test_keys():
        STORAGE.delete_object(bucket_name=STORAGE_CONFIG['storage_bucket'],
                              key=key)


PREFIX = '__pywren.test'
TEST_FILES_URLS = ["http://archive.ics.uci.edu/ml/machine-learning-databases/bag-of-words/vocab.enron.txt",
                   "http://archive.ics.uci.edu/ml/machine-learning-databases/bag-of-words/vocab.kos.txt",
                   "http://archive.ics.uci.edu/ml/machine-learning-databases/bag-of-words/vocab.nips.txt",
                   "http://archive.ics.uci.edu/ml/machine-learning-databases/bag-of-words/vocab.nytimes.txt",
                   "http://archive.ics.uci.edu/ml/machine-learning-databases/bag-of-words/vocab.pubmed.txt"]


def hello_world(param):
    return "Hello World!"


def simple_map_function(x, y):
    return x + y


def simple_reduce_function(results):
    total = 0
    for map_result in results:
        total = total + map_result
    return total


class TestPywren(unittest.TestCase):

    def test_call_async(self):
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.call_async(hello_world, "")
        result = pw.get_result()
        self.assertEqual(result, "Hello World!")

        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.call_async(simple_map_function, [4, 6])
        result = pw.get_result()
        self.assertEqual(result, 10)

        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.call_async(simple_map_function, {'x': 2, 'y': 8})
        result = pw.get_result()
        self.assertEqual(result, 10)

    def test_map(self):
        iterdata = [[1, 1], [2, 2], [3, 3], [4, 4]]
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map(simple_map_function, iterdata)
        result = pw.get_result()
        self.assertEqual(result, [2, 4, 6, 8])

    def test_map_reduce(self):
        iterdata = [[1, 1], [2, 2], [3, 3], [4, 4]]
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(simple_map_function, iterdata, simple_reduce_function)
        result = pw.get_result()
        self.assertEqual(result, 20)

    def test_multiple_executions(self):
        pw = pywren.ibm_cf_executor(config=CONFIG)
        iterdata = [[1, 1], [2, 2]]
        pw.map(simple_map_function, iterdata)
        iterdata = [[3, 3], [4, 4]]
        pw.map(simple_map_function, iterdata)
        result = pw.get_result()
        self.assertEqual(result, [2, 4, 6, 8])

        iterdata = [[1, 1], [2, 2]]
        pw.map(simple_map_function, iterdata)
        result = pw.get_result()
        self.assertEqual(result, [2, 4])

        iterdata = [[1, 1], [2, 2]]
        futures1 = pw.map(simple_map_function, iterdata)
        result1 = pw.get_result(futures=futures1)
        iterdata = [[3, 3], [4, 4]]
        futures2 = pw.map(simple_map_function, iterdata)
        result2 = pw.get_result(futures=futures2)
        self.assertEqual(result1, [2, 4])
        self.assertEqual(result2, [6, 8])


def my_map_function_bucket(bucket, key, data_stream, ibm_cos):
    print('I am processing the object {}'.format(key))
    counter = {}

    data = data_stream.read()
    temp = ibm_cos.get_object(Bucket=bucket, Key=key)['Body'].read()

    for line in data.splitlines():
        for word in line.decode('utf-8').split():
            if word not in counter:
                counter[word] = 1
            else:
                counter[word] += 1

    return counter


def my_map_function_key(key, data_stream, ibm_cos):
    print('I am processing the object {}'.format(key))
    counter = {}

    data = data_stream.read()
    temp = ibm_cos.get_object(Bucket=key.split('/')[0], Key='/'.join(key.split('/')[1:]))['Body'].read()

    for line in data.splitlines():
        for word in line.decode('utf-8').split():
            if word not in counter:
                counter[word] = 1
            else:
                counter[word] += 1

    return counter


def my_map_function_url(url, data_stream):
    print('I am processing the object from {}'.format(url))
    counter = {}

    data = data_stream.read()

    for line in data.splitlines():
        for word in line.decode('utf-8').split():
            if word not in counter:
                counter[word] = 1
            else:
                counter[word] += 1

    return counter


def my_map_function_storage_handler(key_i, bucket_name, ibm_cos):
    print('I am processing the object {}'.format(key_i))
    counter = {}

    data = ibm_cos.get_object(Bucket=bucket_name, Key=key_i)['Body'].read()

    for line in data.splitlines():
        for word in line.decode('utf-8').split():
            if word not in counter:
                counter[word] = 1
            else:
                counter[word] += 1

    return counter


def my_reduce_function(results):
    final_result = 0

    for count in results:
        for word in count:
            final_result += count[word]

    return final_result


class TestPywrenCos(unittest.TestCase):

    def checkResult(self, result):
        result_to_compare = STORAGE.get_object(bucket_name=STORAGE_CONFIG['storage_bucket'],
                                               key=f'{PREFIX}/result')

        if isinstance(result, list):
            total = 0
            for r in result:
                total += r
        else:
            total = result

        self.assertEqual(total, int(result_to_compare))

    def test_map_reduce_cos_bucket(self):
        data_prefix = STORAGE_CONFIG['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_bucket, data_prefix, my_reduce_function)
        result = pw.get_result()
        self.checkResult(result)

    def test_map_reduce_cos_bucket_one_reducer_per_object(self):
        data_prefix = STORAGE_CONFIG['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_bucket, data_prefix, my_reduce_function, reducer_one_per_object=True)
        result = pw.get_result()
        self.checkResult(result)

    def test_map_reduce_cos_key(self):
        bucket_name = STORAGE_CONFIG['storage_bucket']
        iterdata = [bucket_name + '/' + key for key in list_test_keys()]
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_key, iterdata, my_reduce_function)
        result = pw.get_result()
        self.checkResult(result)

    def test_map_reduce_cos_key_one_reducer_per_object(self):
        bucket_name = STORAGE_CONFIG['storage_bucket']
        iterdata = [bucket_name + '/' + key for key in list_test_keys()]
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_key, iterdata, my_reduce_function, reducer_one_per_object=True)
        result = pw.get_result()
        self.checkResult(result)

    def test_map_reduce_url(self):
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_url, TEST_FILES_URLS, my_reduce_function)
        result = pw.get_result()
        self.checkResult(result + 1)

    def test_storage_handler(self):
        iterdata = [[key, STORAGE_CONFIG['storage_bucket']] for key in list_test_keys()]
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_storage_handler, iterdata, my_reduce_function)
        result = pw.get_result()
        self.checkResult(result)

    def test_chunks_bucket(self):
        data_prefix = STORAGE_CONFIG['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_bucket, data_prefix, my_reduce_function, chunk_size=1 * 1024 ** 2)
        result = pw.get_result()
        self.checkResult(result)

    def test_chunks_bucket_one_reducer_per_object(self):
        data_prefix = STORAGE_CONFIG['storage_bucket'] + '/' + PREFIX
        pw = pywren.ibm_cf_executor(config=CONFIG)
        pw.map_reduce(my_map_function_bucket, data_prefix, my_reduce_function, chunk_size=1 * 1024 ** 2,
                      reducer_one_per_object=True)
        result = pw.get_result()
        self.checkResult(result)


def run(config=None):
    global CONFIG
    global STORAGE_CONFIG
    global STORAGE

    CONFIG = wrenconfig.default(config)
    STORAGE_CONFIG = wrenconfig.extract_storage_config(CONFIG)
    internal_storage = InternalStorage(STORAGE_CONFIG)
    STORAGE = internal_storage.storage_handler

    if len(sys.argv) <= 1:
        task = 'full'
    else:
        task = sys.argv[1]

    suite = unittest.TestSuite()
    if task == 'pywren':
        suite.addTest(unittest.makeSuite(TestPywren))
    elif task == 'pywren_cos':
        suite.addTest(unittest.makeSuite(TestPywrenCos))
    elif task == 'full':
        suite.addTest(unittest.makeSuite(TestPywren))
        suite.addTest(unittest.makeSuite(TestPywrenCos))
    elif task == 'test_call_async':
        suite.addTest(TestPywren('test_call_async'))
    elif task == 'test_map':
        suite.addTest(TestPywren('test_map'))
    elif task == 'test_map_reduce':
        suite.addTest(TestPywren('test_map_reduce'))
    elif task == 'test_multiple_executions':
        suite.addTest(TestPywren('test_multiple_executions'))
    elif task == 'test_map_reduce_cos_bucket':
        suite.addTest(TestPywrenCos('test_map_reduce_cos_bucket'))
    elif task == 'test_map_reduce_cos_bucket_one_reducer_per_object':
        suite.addTest(TestPywrenCos('test_map_reduce_cos_bucket_one_reducer_per_object'))
    elif task == 'test_map_reduce_cos_key':
        suite.addTest(TestPywrenCos('test_map_reduce_cos_key'))
    elif task == 'test_map_reduce_cos_key_one_reducer_per_object':
        suite.addTest(TestPywrenCos('test_map_reduce_cos_key_one_reducer_per_object'))
    elif task == 'test_map_reduce_url':
        suite.addTest(TestPywrenCos('test_map_reduce_url'))
    elif task == 'test_storage_handler':
        suite.addTest(TestPywrenCos('test_storage_handler'))
    elif task == 'test_chunks_bucket':
        suite.addTest(TestPywrenCos('test_chunks_bucket'))
    elif task == 'test_chunks_bucket_one_reducer_per_object':
        suite.addTest(TestPywrenCos('test_chunks_bucket_one_reducer_per_object'))
    else:
        print('Unknown Command... use: "init", "pywren", "pywren_cos", "clean" or a test function name.')
        sys.exit()

    initTests()
    runner = unittest.TextTestRunner()
    runner.run(suite)
    cleanTests()


if __name__ == '__main__':
    run()



from pywren_ibm_cloud.runtime.deploy_utils import create_runtime
from pywren_ibm_cloud.runtime.deploy_utils import build_runtime
from pywren_ibm_cloud.runtime.deploy_utils import update_runtime
from pywren_ibm_cloud.runtime.deploy_utils import delete_runtime
from pywren_ibm_cloud.runtime.deploy_utils import clean_runtimes

from pywren_ibm_cloud.runtime.runtime import select_runtime



#!/usr/bin/env python3
from pywren_ibm_cloud.runtime import create_runtime, update_runtime, build_runtime, delete_runtime, clean_runtimes
import logging
import click
import os

logging.basicConfig(level=logging.DEBUG)
os.environ["PYWREN_LOG_LEVEL"] = 'DEBUG'


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.command('create')
@click.argument('image_name')
@click.option('--memory', default=None, help='memory used by the runtime', type=int)
def create(image_name, memory):
    create_runtime(image_name, memory=memory)


@cli.command('build')
@click.argument('image_name')
def build(image_name):
    build_runtime(image_name)


@cli.command('update')
@click.argument('image_name')
def update(image_name):
    update_runtime(image_name)


@cli.command('delete')
@click.argument('image_name')
def delete(image_name):
    delete_runtime(image_name)


@cli.command('clean')
def clean():
    clean_runtimes()


if __name__ == "__main__":
    cli()



#
# (C) Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import ibm_boto3
import ibm_botocore
from datetime import datetime
from ibm_botocore.credentials import DefaultTokenManager
from ...utils import sizeof_fmt, is_cf_cluster
from ..exceptions import StorageNoSuchKeyError

logging.getLogger('ibm_boto3').setLevel(logging.CRITICAL)
logging.getLogger('ibm_botocore').setLevel(logging.CRITICAL)
logging.getLogger('urllib3').setLevel(logging.CRITICAL)
logger = logging.getLogger(__name__)


class IbmCosStorageBackend:
    """
    A wrap-up around IBM COS ibm_boto3 APIs.
    """

    def __init__(self, ibm_cos_config):
        self.is_cf_cluster = is_cf_cluster()
        self.ibm_cos_config = ibm_cos_config
        iam_config = ibm_cos_config['ibm_iam']

        service_endpoint = ibm_cos_config.get('endpoint').replace('http:', 'https:')
        if self.is_cf_cluster and 'private_endpoint' in ibm_cos_config:
            service_endpoint = ibm_cos_config.get('private_endpoint')
            if 'api_key' in ibm_cos_config:
                service_endpoint = service_endpoint.replace('http:', 'https:')

        ibm_auth_endpoint = iam_config['ibm_auth_endpoint']
        logger.debug("Set IBM COS Endpoint to {}".format(service_endpoint))
        logger.debug("Set IBM IAM Auth Endpoint to {}".format(ibm_auth_endpoint))

        api_key = None
        if 'api_key' in ibm_cos_config:
            api_key = ibm_cos_config.get('api_key')
            api_key_type = 'COS'
        elif 'api_key' in iam_config:
            api_key = iam_config.get('api_key')
            api_key_type = 'IAM'

        if {'secret_key', 'access_key'} <= set(ibm_cos_config):
            logger.debug("IBM COS using access_key and secret_key")
            access_key = ibm_cos_config.get('access_key')
            secret_key = ibm_cos_config.get('secret_key')
            client_config = ibm_botocore.client.Config(max_pool_connections=128,
                                                       user_agent_extra='pywren-ibm-cloud',
                                                       connect_timeout=1)
            self.cos_client = ibm_boto3.client('s3',
                                               aws_access_key_id=access_key,
                                               aws_secret_access_key=secret_key,
                                               config=client_config,
                                               endpoint_url=service_endpoint)

        elif api_key is not None:
            client_config = ibm_botocore.client.Config(signature_version='oauth',
                                                       max_pool_connections=128,
                                                       user_agent_extra='pywren-ibm-cloud',
                                                       connect_timeout=1)
            token_manager = DefaultTokenManager(api_key_id=api_key, auth_endpoint=ibm_auth_endpoint)

            if 'token' not in ibm_cos_config:
                logger.debug("IBM COS: Using {} api_key - Requesting new token".format(api_key_type))
                ibm_cos_config['token'] = token_manager.get_token()
                ibm_cos_config['token_expiry_time'] = token_manager._expiry_time.strftime('%Y-%m-%d %H:%M:%S.%f%z')

            else:
                logger.debug("IBM COS: Using {} api_key - Reusing token".format(api_key_type))
                token_manager._token = ibm_cos_config['token']
                expiry_time = ibm_cos_config['token_expiry_time']
                token_manager._expiry_time = datetime.strptime(expiry_time, '%Y-%m-%d %H:%M:%S.%f%z')

                if token_manager._is_expired() and not self.is_cf_cluster:
                    # Only request new token on client machine
                    logger.debug("IBM COS: Using {} api_key - Token expired, requesting new token".format(api_key_type))
                    ibm_cos_config['token'] = token_manager.get_token()
                    ibm_cos_config['token_expiry_time'] = token_manager._expiry_time.strftime('%Y-%m-%d %H:%M:%S.%f%z')

            self.cos_client = ibm_boto3.client('s3', token_manager=token_manager,
                                               config=client_config,
                                               endpoint_url=service_endpoint)

    def get_client(self):
        """
        Get ibm_boto3 client.
        :return: ibm_boto3 client
        """
        return self.cos_client

    def put_object(self, bucket_name, key, data):
        """
        Put an object in COS. Override the object if the key already exists.
        :param key: key of the object.
        :param data: data of the object
        :type data: str/bytes
        :return: None
        """
        try:
            res = self.cos_client.put_object(Bucket=bucket_name, Key=key, Body=data)
            status = 'OK' if res['ResponseMetadata']['HTTPStatusCode'] == 200 else 'Error'
            try:
                logger.debug('PUT Object {} - Size: {} - {}'.format(key, sizeof_fmt(len(data)), status))
            except Exception:
                logger.debug('PUT Object {} {}'.format(key, status))
        except ibm_botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "NoSuchKey":
                raise StorageNoSuchKeyError(key)
            else:
                raise e

    def get_object(self, bucket_name, key, stream=False, extra_get_args={}):
        """
        Get object from COS with a key. Throws StorageNoSuchKeyError if the given key does not exist.
        :param key: key of the object
        :return: Data of the object
        :rtype: str/bytes
        """
        try:
            r = self.cos_client.get_object(Bucket=bucket_name, Key=key, **extra_get_args)
            if stream:
                data = r['Body']
            else:
                data = r['Body'].read()
            return data
        except ibm_botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "NoSuchKey":
                raise StorageNoSuchKeyError(key)
            else:
                raise e

    def head_object(self, bucket_name, key):
        """
        Head object from COS with a key. Throws StorageNoSuchKeyError if the given key does not exist.
        :param key: key of the object
        :return: Data of the object
        :rtype: str/bytes
        """
        try:
            metadata = self.cos_client.head_object(Bucket=bucket_name, Key=key)
            return metadata['ResponseMetadata']['HTTPHeaders']
        except ibm_botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                raise StorageNoSuchKeyError(key)
            else:
                raise e

    def delete_object(self, bucket_name, key):
        """
        Delete an object from storage.
        :param bucket: bucket name
        :param key: data key
        """
        return self.cos_client.delete_object(Bucket=bucket_name, Key=key)

    def delete_objects(self, bucket_name, key_list):
        """
        Delete a list of objects from storage.
        :param bucket: bucket name
        :param key_list: list of keys
        """
        result = []
        max_keys_num = 1000
        for i in range(0, len(key_list), max_keys_num):
            delete_keys = {'Objects': []}
            delete_keys['Objects'] = [{'Key': k} for k in key_list[i:i+max_keys_num]]
            result.append(self.cos_client.delete_objects(Bucket=bucket_name, Delete=delete_keys))
        return result

    def bucket_exists(self, bucket_name):
        """
        Head bucket from COS with a name. Throws StorageNoSuchKeyError if the given bucket does not exist.
        :param bucket_name: name of the bucket
        :return: Data of the object
        :rtype: str/bytes
        """
        try:
            self.cos_client.head_bucket(Bucket=bucket_name)
        except ibm_botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                raise StorageNoSuchKeyError(bucket_name)
            else:
                raise e

    def list_objects(self, bucket_name, prefix=None):
        paginator = self.cos_client.get_paginator('list_objects_v2')
        try:
            if (prefix is not None):
                page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
            else:
                page_iterator = paginator.paginate(Bucket=bucket_name)

            object_list = []
            for page in page_iterator:
                if 'Contents' in page:
                    for item in page['Contents']:
                        object_list.append(item)
            return object_list
        except ibm_botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                raise StorageNoSuchKeyError(bucket_name)
            else:
                raise e

    def list_keys_with_prefix(self, bucket_name, prefix):
        """
        Return a list of keys for the given prefix.
        :param prefix: Prefix to filter object names.
        :return: List of keys in bucket that match the given prefix.
        :rtype: list of str
        """
        if not prefix:
            prefix = ''
        paginator = self.cos_client.get_paginator('list_objects_v2')
        operation_parameters = {'Bucket': bucket_name,
                                'Prefix': prefix}
        page_iterator = paginator.paginate(**operation_parameters)

        key_list = []
        for page in page_iterator:
            if 'Contents' in page:
                for item in page['Contents']:
                    key_list.append(item['Key'])

        return key_list



import os
import sys
import logging
from pywren_ibm_cloud.compute import Compute
from pywren_ibm_cloud.utils import version_str
from pywren_ibm_cloud.runtime import create_runtime
from pywren_ibm_cloud.wrenconfig import extract_compute_config

logger = logging.getLogger(__name__)


def select_runtime(config, internal_storage, executor_id, runtime_name, runtime_memory):
    """
    Auxiliary method that gets the runtime metadata from the storage. This metadata contains the preinstalled
    python modules needed to serialize the local function.  If the .metadata file does not exists in the storage,
    this means that the runtime is not installed, so this method will proceed to install it.
    """
    log_level = os.getenv('PYWREN_LOG_LEVEL')
    compute_config = extract_compute_config(config)
    internal_compute = Compute(compute_config)

    log_msg = 'ExecutorID {} - Selected Runtime: {} - {}MB'.format(executor_id, runtime_name, runtime_memory)
    logger.info(log_msg)
    if not log_level:
        print(log_msg, end=' ')

    runtime_key = internal_compute.get_runtime_key(runtime_name, runtime_memory)
    try:
        runtime_meta = internal_storage.get_runtime_info(runtime_key)
        if not log_level:
            print()
    except Exception:
        logger.debug('ExecutorID {} - Runtime {} with {}MB is not yet installed'.format(executor_id, runtime_name, runtime_memory))
        if not log_level:
            print('(Installing...)')
        create_runtime(runtime_name, memory=runtime_memory, config=config)
        runtime_meta = internal_storage.get_runtime_info(runtime_key)

    if not _runtime_valid(runtime_meta):
        raise Exception(("The indicated runtime: {} "
                         "is not appropriate for this Python version.")
                        .format(runtime_name))

    return runtime_meta['preinstalls']


def _runtime_valid(runtime_meta):
    """
    Basic checks
    """
    this_version_str = version_str(sys.version_info)
    return this_version_str == runtime_meta['python_ver']



#
# (C) Copyright IBM Corp. 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import json
import time
import shutil
import pickle
import logging
import inspect
import numpy as np
from multiprocessing import Process
from distutils.util import strtobool
from pywren_ibm_cloud.storage import InternalStorage
from pywren_ibm_cloud.future import ResponseFuture
from pywren_ibm_cloud.libs.tblib import pickling_support
from pywren_ibm_cloud.utils import sizeof_fmt, b64str_to_bytes
from pywren_ibm_cloud.wrenconfig import extract_storage_config
from pywren_ibm_cloud.utils import get_current_memory_usage
from pywren_ibm_cloud.storage.backends.ibm_cos import IbmCosStorageBackend
from pywren_ibm_cloud.storage.backends.swift import SwiftStorageBackend
from pywren_ibm_cloud.logging_config import ibm_cf_logging_config

pickling_support.install()
logger = logging.getLogger('jobrunner')


class stats:

    def __init__(self, stats_filename):
        self.stats_filename = stats_filename
        self.stats_fid = open(stats_filename, 'w')

    def write(self, key, value):
        self.stats_fid.write("{} {}\n".format(key, value))
        self.stats_fid.flush()

    def __del__(self):
        self.stats_fid.close()


class jobrunner(Process):

    def __init__(self, jr_config, result_queue):
        super().__init__()
        start_time = time.time()
        self.config = jr_config
        log_level = self.config['log_level']
        self.result_queue = result_queue
        ibm_cf_logging_config(log_level)
        self.stats = stats(self.config['stats_filename'])
        self.stats.write('jobrunner_start', start_time)
        pw_config = json.loads(os.environ.get('PYWREN_CONFIG'))
        self.storage_config = extract_storage_config(pw_config)

        if 'SHOW_MEMORY_USAGE' in os.environ:
            self.show_memory = eval(os.environ['SHOW_MEMORY_USAGE'])
        else:
            self.show_memory = False

        self.func_key = self.config['func_key']
        self.data_key = self.config['data_key']
        self.data_byte_range = self.config['data_byte_range']
        self.output_key = self.config['output_key']

    def _get_function_and_modules(self):
        """
        Gets and unpickles function and modules from storage
        """
        logger.debug("Getting function and modules")
        func_download_time_t1 = time.time()
        func_obj = self.internal_storage.get_func(self.func_key)
        loaded_func_all = pickle.loads(func_obj)
        func_download_time_t2 = time.time()
        self.stats.write('func_download_time', round(func_download_time_t2-func_download_time_t1, 8))
        logger.debug("Finished getting Function and modules")

        return loaded_func_all

    def _save_modules(self, module_data):
        """
        Save modules, before we unpickle actual function
        """
        logger.debug("Writing Function dependencies to local disk")
        PYTHON_MODULE_PATH = self.config['python_module_path']
        shutil.rmtree(PYTHON_MODULE_PATH, True)  # delete old modules
        os.mkdir(PYTHON_MODULE_PATH)
        sys.path.append(PYTHON_MODULE_PATH)

        for m_filename, m_data in module_data.items():
            m_path = os.path.dirname(m_filename)

            if len(m_path) > 0 and m_path[0] == "/":
                m_path = m_path[1:]
            to_make = os.path.join(PYTHON_MODULE_PATH, m_path)
            try:
                os.makedirs(to_make)
            except OSError as e:
                if e.errno == 17:
                    pass
                else:
                    raise e
            full_filename = os.path.join(to_make, os.path.basename(m_filename))

            with open(full_filename, 'wb') as fid:
                fid.write(b64str_to_bytes(m_data))

        #logger.info("Finished writing {} module files".format(len(loaded_func_all['module_data'])))
        #logger.debug(subprocess.check_output("find {}".format(PYTHON_MODULE_PATH), shell=True))
        #logger.debug(subprocess.check_output("find {}".format(os.getcwd()), shell=True))
        logger.debug("Finished writing Function dependencies")

    def _unpickle_function(self, pickled_func):
        """
        Unpickle function; it will expect modules to be there
        """
        logger.debug("Unpickle Function")
        loaded_func = pickle.loads(pickled_func)
        logger.debug("Finished Function unpickle")

        return loaded_func

    def _load_data(self):
        extra_get_args = {}
        if self.data_byte_range is not None:
            range_str = 'bytes={}-{}'.format(*self.data_byte_range)
            extra_get_args['Range'] = range_str

        logger.debug("Getting function data")
        data_download_time_t1 = time.time()
        data_obj = self.internal_storage.get_data(self.data_key, extra_get_args=extra_get_args)
        logger.debug("Finished getting Function data")
        logger.debug("Unpickle Function data")
        loaded_data = pickle.loads(data_obj)
        logger.debug("Finished unpickle Function data")
        data_download_time_t2 = time.time()
        self.stats.write('data_download_time', round(data_download_time_t2-data_download_time_t1, 8))

        return loaded_data

    def _create_storage_clients(self, function, data):
        # Verify storage parameters - Create clients
        func_sig = inspect.signature(function)

        if 'ibm_cos' in func_sig.parameters:
            ibm_boto3_client = IbmCosStorageBackend(self.storage_config['ibm_cos']).get_client()
            data['ibm_cos'] = ibm_boto3_client

        if 'swift' in func_sig.parameters:
            swift_client = SwiftStorageBackend(self.storage_config['swift'])
            data['swift'] = swift_client

        if 'internal_storage' in func_sig.parameters:
            data['internal_storage'] = self.internal_storage

        return data

    def run(self):
        """
        Runs the function
        """
        logger.info("Started")
        # initial output file in case job fails
        result = None
        exception = False
        try:
            self.internal_storage = InternalStorage(self.storage_config)

            loaded_func_all = self._get_function_and_modules()
            self._save_modules(loaded_func_all['module_data'])
            function = self._unpickle_function(loaded_func_all['func'])
            data = self._load_data()
            data = self._create_storage_clients(function, data)

            if self.show_memory:
                logger.debug("Memory usage before call the function: {}".format(get_current_memory_usage()))

            logger.info("Function: Going to execute '{}()'".format(str(function.__name__)))
            print('---------------------- FUNCTION LOG ----------------------', flush=True)
            func_exec_time_t1 = time.time()
            result = function(**data)
            func_exec_time_t2 = time.time()
            print('----------------------------------------------------------', flush=True)
            logger.info("Function: Success execution")

            if self.show_memory:
                logger.debug("Memory usage after call the function: {}".format(get_current_memory_usage()))

            self.stats.write('function_exec_time', round(func_exec_time_t2-func_exec_time_t1, 8))

            # Check for new futures
            if result is not None:
                self.stats.write("result", True)
                if isinstance(result, ResponseFuture):
                    callgroup_id = result.callgroup_id
                    self.stats.write('new_futures', '{}/{}'.format(callgroup_id, 1))
                elif type(result) == list and len(result) > 0 and isinstance(result[0], ResponseFuture):
                    callgroup_id = result[0].callgroup_id
                    self.stats.write('new_futures', '{}/{}'.format(callgroup_id, len(result)))
                else:
                    self.stats.write('new_futures', '{}/{}'.format(None, 0))

                logger.debug("Pickling result")
                output_dict = {'result': result}
                pickled_output = pickle.dumps(output_dict)

                if self.show_memory:
                    logger.debug("Memory usage after output serialization: {}".format(get_current_memory_usage()))
            else:
                logger.debug("No result to store")
                self.stats.write("result", False)

        except Exception as e:
            exception = True
            self.stats.write("exception", True)
            print('----------------------- EXCEPTION !-----------------------')
            logger.error("There was an exception: {}".format(str(e)))
            print('----------------------------------------------------------', flush=True)

            if self.show_memory:
                logger.debug("Memory usage after call the function: {}".format(get_current_memory_usage()))

            try:
                logger.debug("Pickling exception")
                pickled_exc = pickle.dumps(sys.exc_info())
                pickle.loads(pickled_exc)  # this is just to make sure they can be unpickled
                self.stats.write("exc_info", str(pickled_exc))

            except Exception as pickle_exception:
                # Shockingly often, modules like subprocess don't properly
                # call the base Exception.__init__, which results in them
                # being unpickleable. As a result, we actually wrap this in a try/catch block
                # and more-carefully handle the exceptions if any part of this save / test-reload
                # fails
                logger.debug("Failed pickling exception: {}".format(str(pickle_exception)))
                self.stats.write("exc_pickle_fail", True)
                exc_type, exc_value, exc_traceback = sys.exc_info()
                pickled_exc = pickle.dumps({'exc_type': str(exc_type),
                                            'exc_value': str(exc_value),
                                            'exc_traceback': exc_traceback,
                                            'pickle_exception': pickle_exception})
                pickle.loads(pickled_exc)  # this is just to make sure they can be unpickled
                self.stats.write("exc_info", str(pickled_exc))
        finally:
            store_result = strtobool(os.environ.get('STORE_RESULT', 'True'))
            if result is not None and store_result and not exception:
                output_upload_timestamp_t1 = time.time()
                logger.info("Storing function result - output.pickle - Size: {}".format(sizeof_fmt(len(pickled_output))))
                self.internal_storage.put_data(self.output_key, pickled_output)
                output_upload_timestamp_t2 = time.time()
                self.stats.write("output_upload_time", round(output_upload_timestamp_t2 - output_upload_timestamp_t1, 8))
            self.result_queue.put("Finished")
            logger.info("Finished")



import os
import sys
import time
import enum
import json
import signal
import logging
import traceback
from pywren_ibm_cloud import wrenconfig
from pywren_ibm_cloud.invoker import Invoker
from pywren_ibm_cloud.storage import InternalStorage
from pywren_ibm_cloud.future import FunctionException
from pywren_ibm_cloud.wait import wait, ALL_COMPLETED
from pywren_ibm_cloud.storage.cleaner import clean_os_bucket
from pywren_ibm_cloud.logging_config import default_logging_config
from pywren_ibm_cloud.job import create_call_async_job, create_map_job, create_reduce_job
from pywren_ibm_cloud.utils import timeout_handler, is_notebook, is_unix_system, is_cf_cluster, create_executor_id

logger = logging.getLogger(__name__)


class ExecutorState(enum.Enum):
    new = 1
    running = 2
    ready = 3
    success = 4
    error = 5
    finished = 6


class ibm_cf_executor:

    def __init__(self, config=None, runtime=None, runtime_memory=None, log_level=None, rabbitmq_monitor=False):
        """
        Initialize and return a ServerlessExecutor class.

        :param config: Settings passed in here will override those in config file. Default None.
        :param runtime: Runtime name to use. Default None.
        :param runtime_memory: memory to use in the runtime
        :param log_level: log level to use during the execution
        :param rabbitmq_monitor: use rabbitmq as monitoring system
        :return `ServerlessExecutor` object.

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
        """
        self.start_time = time.time()
        self._state = ExecutorState.new
        self.config = wrenconfig.default(config)
        self.is_cf_cluster = is_cf_cluster()
        self.data_cleaner = self.config['pywren']['data_cleaner']

        # Overwrite runtime variables
        if runtime:
            self.config['pywren']['runtime'] = runtime
        if runtime_memory:
            self.config['pywren']['runtime_memory'] = int(runtime_memory)

        # Log level Configuration
        self.log_level = log_level
        if not self.log_level:
            if(logger.getEffectiveLevel() != logging.WARNING):
                self.log_level = logging.getLevelName(logger.getEffectiveLevel())
        if self.log_level:
            os.environ["PYWREN_LOG_LEVEL"] = self.log_level
            if not self.is_cf_cluster:
                default_logging_config(self.log_level)

        if 'PYWREN_EXECUTOR_ID' in os.environ:
            self.executor_id = os.environ['PYWREN_EXECUTOR_ID']
        else:
            self.executor_id = create_executor_id()
        logger.debug('ServerlessExecutor created with ID: {}'.format(self.executor_id))

        # RabbitMQ monitor configuration
        self.rabbitmq_monitor = rabbitmq_monitor
        if self.rabbitmq_monitor:
            if self.config['rabbitmq']['amqp_url']:
                os.environ["PYWREN_RABBITMQ_MONITOR"] = 'True'
            else:
                self.rabbitmq_monitor = False
        else:
            self.config['rabbitmq']['amqp_url'] = None

        storage_config = wrenconfig.extract_storage_config(self.config)
        self.internal_storage = InternalStorage(storage_config)
        self.invoker = Invoker(self.config, self.executor_id)

        self.executor_futures = []
        self.futures = []

    def call_async(self, func, data, extra_env=None, extra_meta=None, runtime_memory=None, timeout=wrenconfig.RUNTIME_TIMEOUT):
        """
        For running one function execution asynchronously
        :param func: the function to map over the data
        :param data: input data
        :param extra_env: Additional environment variables for action environment. Default None.
        :param extra_meta: Additional metadata to pass to action. Default None.

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> future = pw.call_async(foo, data)
        """

        if self._state == ExecutorState.finished:
            raise Exception('You cannot run pw.call_async() in the current state,'
                            ' create a new pywren.ibm_cf_executor() instance.')

        job = create_call_async_job(self.config, self.internal_storage, self.executor_id,
                                    func, data, extra_env, extra_meta, runtime_memory, timeout)
        future = self.invoker.run(job)

        self.futures.extend(future)
        self._state = ExecutorState.running

        return future

    def map(self, map_function, map_iterdata, extra_env=None, extra_meta=None, runtime_memory=None,
            chunk_size=None, remote_invocation=False, timeout=wrenconfig.RUNTIME_TIMEOUT,
            remote_invocation_groups=None, invoke_pool_threads=500, overwrite_invoke_args=None, exclude_modules=None):
        """
        :param func: the function to map over the data
        :param iterdata: An iterable of input data
        :param extra_env: Additional environment variables for action environment. Default None.
        :param extra_meta: Additional metadata to pass to action. Default None.
        :param chunk_size: the size of the data chunks. 'None' for processing the whole file in one map
        :param remote_invocation: Enable or disable remote_invocayion mechanism. Default 'False'
        :param timeout: Time that the functions have to complete their execution before raising a timeout.
        :param data_type: the type of the data. Now allowed: None (files with newline) and csv.
        :param invoke_pool_threads: Number of threads to use to invoke.
        :param data_all_as_one: upload the data as a single object. Default True
        :param overwrite_invoke_args: Overwrite other args. Mainly used for testing.
        :param exclude_modules: Explicitly keep these modules from pickled dependencies.
        :return: A list with size `len(iterdata)` of futures for each job
        :rtype: list of futures.

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> futures = pw.map(foo, data_list)
        """
        if self._state == ExecutorState.finished:
            raise Exception('You cannot run map() in the current state.'
                            ' Create a new ibm_cf_executor() instance.')
        job, unused_ppo = create_map_job(self.config, self.internal_storage, self.executor_id,
                                         map_function=map_function, iterdata=map_iterdata,
                                         extra_env=extra_env, extra_meta=extra_meta,
                                         obj_chunk_size=chunk_size, runtime_memory=runtime_memory,
                                         remote_invocation=remote_invocation,
                                         remote_invocation_groups=remote_invocation_groups,
                                         invoke_pool_threads=invoke_pool_threads,
                                         exclude_modules=exclude_modules,
                                         is_cf_cluster=self.is_cf_cluster,
                                         overwrite_invoke_args=overwrite_invoke_args,
                                         runtime_timeout=timeout)
        map_futures = self.invoker.run(job)

        self.futures.extend(map_futures)
        self._state = ExecutorState.running

        if len(map_futures) == 1:
            return map_futures[0]
        return map_futures

    def map_reduce(self, map_function, map_iterdata, reduce_function, extra_env=None,
                   map_runtime_memory=None, reduce_runtime_memory=None,
                   extra_meta=None, chunk_size=None, remote_invocation=False,
                   remote_invocation_groups=None, timeout=wrenconfig.RUNTIME_TIMEOUT,
                   reducer_one_per_object=False, reducer_wait_local=False,
                   invoke_pool_threads=500, overwrite_invoke_args=None,
                   exclude_modules=None):
        """
        Map the map_function over the data and apply the reduce_function across all futures.
        This method is executed all within CF.
        :param map_function: the function to map over the data
        :param map_iterdata:  the function to reduce over the futures
        :param reduce_function:  the function to reduce over the futures
        :param extra_env: Additional environment variables for action environment. Default None.
        :param extra_meta: Additional metadata to pass to action. Default None.
        :param chunk_size: the size of the data chunks. 'None' for processing the whole file in one map
        :param remote_invocation: Enable or disable remote_invocayion mechanism. Default 'False'
        :param timeout: Time that the functions have to complete their execution before raising a timeout.
        :param data_type: the type of the data. Now allowed: None (files with newline) and csv.
        :param reducer_one_per_object: Set one reducer per object after running the partitioner
        :param reducer_wait_local: Wait for results locally
        :param invoke_pool_threads: Number of threads to use to invoke.
        :param data_all_as_one: upload the data as a single object. Default True
        :param overwrite_invoke_args: Overwrite other args. Mainly used for testing.
        :param exclude_modules: Explicitly keep these modules from pickled dependencies.
        :return: A list with size `len(map_iterdata)` of futures for each job

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> pw.map_reduce(foo, map_data_list, bar)
        """

        if self._state == ExecutorState.finished:
            raise Exception('You cannot run map_reduce() in the current state.'
                            ' Create a new ibm_cf_executor() instance.')

        job, parts_per_object = create_map_job(self.config, self.internal_storage, self.executor_id,
                                               map_function=map_function, iterdata=map_iterdata,
                                               extra_env=extra_env, extra_meta=extra_meta,
                                               obj_chunk_size=chunk_size, runtime_memory=map_runtime_memory,
                                               remote_invocation=remote_invocation,
                                               remote_invocation_groups=remote_invocation_groups,
                                               invoke_pool_threads=invoke_pool_threads,
                                               exclude_modules=exclude_modules,
                                               is_cf_cluster=self.is_cf_cluster,
                                               overwrite_invoke_args=overwrite_invoke_args,
                                               runtime_timeout=timeout)
        map_futures = self.invoker.run(job)

        self._state = ExecutorState.running
        if reducer_wait_local:
            self.monitor(futures=map_futures)

        job = create_reduce_job(self.config, self.internal_storage, self.executor_id, reduce_function,
                                reduce_runtime_memory, map_futures, parts_per_object, reducer_one_per_object,
                                extra_env, extra_meta)
        reduce_future = self.invoker.run(job)

        for f in map_futures:
            f.produce_output = False
        futures = map_futures + reduce_future
        self.futures.extend(futures)

        return futures

    def monitor(self, futures=None, throw_except=True, return_when=ALL_COMPLETED,
                download_results=False, timeout=wrenconfig.RUNTIME_TIMEOUT,
                THREADPOOL_SIZE=128, WAIT_DUR_SEC=1):
        """
        Wait for the Future instances `fs` to complete. Returns a 2-tuple of
        lists. The first list contains the futures that completed
        (finished or cancelled) before the wait completed. The second
        contains uncompleted futures.
        :param futures: Futures list. Default None
        :param throw_except: Re-raise exception if call raised. Default True.
        :param return_when: One of `ALL_COMPLETED`, `ANY_COMPLETED`, `ALWAYS`
        :param download_results: Download results. Default false (Only download statuses)
        :param timeout: Timeout of waiting for results.
        :param THREADPOOL_SIZE: Number of threads to use. Default 64
        :param WAIT_DUR_SEC: Time interval between each check.
        :return: `(fs_done, fs_notdone)`
            where `fs_done` is a list of futures that have completed
            and `fs_notdone` is a list of futures that have not completed.
        :rtype: 2-tuple of lists

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> pw.map(foo, data_list)
          >>> dones, not_dones = pw.monitor()
          >>> # not_dones should be an empty list.
          >>> results = [f.result() for f in dones]
        """
        if futures:
            # Ensure futures is a list
            if type(futures) != list:
                ftrs = [futures]
            else:
                ftrs = futures
        else:
            # In this case self.futures is always a list
            ftrs = self.futures

        if not ftrs:
            raise Exception('You must run call_async(), map() or map_reduce()'
                            ' before calling get_result() method')

        rabbit_amqp_url = None
        if self.rabbitmq_monitor:
            rabbit_amqp_url = self.config['rabbitmq'].get('amqp_url')
        if rabbit_amqp_url and not download_results:
            logger.info('Going to use RabbitMQ to monitor function activations')
            logging.getLogger('pika').setLevel(logging.WARNING)
        if download_results:
            msg = 'ExecutorID {} - Getting results...'.format(self.executor_id)
        else:
            msg = 'ExecutorID {} - Waiting for functions to complete...'.format(self.executor_id)
        logger.info(msg)
        if not self.log_level and self._state == ExecutorState.running:
            print(msg)

        if is_unix_system():
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(timeout)

        pbar = None
        if not self.is_cf_cluster and self._state == ExecutorState.running \
           and not self.log_level:
            from tqdm.auto import tqdm
            if is_notebook():
                pbar = tqdm(bar_format='{n}/|/ {n_fmt}/{total_fmt}', total=len(ftrs))  # ncols=800
            else:
                print()
                pbar = tqdm(bar_format='  {l_bar}{bar}| {n_fmt}/{total_fmt}  ', total=len(ftrs), disable=False)

        try:
            wait(ftrs, self.executor_id, self.internal_storage, download_results=download_results,
                 throw_except=throw_except, return_when=return_when, rabbit_amqp_url=rabbit_amqp_url,
                 pbar=pbar, THREADPOOL_SIZE=THREADPOOL_SIZE, WAIT_DUR_SEC=WAIT_DUR_SEC)

        except FunctionException as e:
            if is_unix_system():
                signal.alarm(0)
            if pbar:
                pbar.close()
            logger.info(e.msg)
            if not is_notebook():
                print()
            if not self.log_level:
                print(e.msg)
            if e.exc_msg:
                print('--> Exception: ' + e.exc_msg)
            else:
                print()
                traceback.print_exception(*e.exception)
            sys.exit()

        except TimeoutError:
            if download_results:
                not_dones_activation_ids = [f.activation_id for f in ftrs if not f.done and not (f.ready and not f.produce_output)]
            else:
                not_dones_activation_ids = [f.activation_id for f in ftrs if not f.ready]
            msg = ('ExecutorID {} - Raised timeout of {} seconds waiting for results '
                   '\nActivations not done: {}'.format(self.executor_id, timeout, not_dones_activation_ids))
            self._state = ExecutorState.error

        except KeyboardInterrupt:
            if download_results:
                not_dones_activation_ids = [f.activation_id for f in ftrs if not f.done and not (f.ready and not f.produce_output)]
            else:
                not_dones_activation_ids = [f.activation_id for f in ftrs if not f.ready]
            msg = 'ExecutorID {} - Cancelled  \nActivations not done: {}'.format(self.executor_id, not_dones_activation_ids)
            self._state = ExecutorState.error

        finally:
            if is_unix_system():
                signal.alarm(0)
            if pbar:
                pbar.close()
                if not is_notebook():
                    print()
            if self._state == ExecutorState.error:
                logger.info(msg)
                if not self.log_level:
                    print(msg)
            if download_results and self.data_cleaner and not self.is_cf_cluster:
                self.clean()

        if download_results:
            fs_dones = [f for f in ftrs if f.done]
            fs_notdones = [f for f in ftrs if not f.done]
        else:
            fs_dones = [f for f in ftrs if f.ready]
            fs_notdones = [f for f in ftrs if not f.ready]

        self._state = ExecutorState.ready

        return fs_dones, fs_notdones

    def get_result(self, futures=None, throw_except=True, timeout=wrenconfig.RUNTIME_TIMEOUT,
                   THREADPOOL_SIZE=64, WAIT_DUR_SEC=1):
        """
        For getting results
        :param futures: Futures list. Default None
        :param throw_except: Reraise exception if call raised. Default True.
        :param verbose: Shows some information prints. Default False
        :param timeout: Timeout for waiting for results.
        :param THREADPOOL_SIZE: Number of threads to use. Default 64
        :param WAIT_DUR_SEC: Time interval between each check.
        :return: The result of the future/s

        Usage
          >>> import pywren_ibm_cloud as pywren
          >>> pw = pywren.ibm_cf_executor()
          >>> pw.map(foo, data)
          >>> results = pw.get_result()
        """
        fs_dones, unused_fs_notdones = self.monitor(futures=futures, throw_except=throw_except,
                                                    timeout=timeout, download_results=True,
                                                    THREADPOOL_SIZE=THREADPOOL_SIZE,
                                                    WAIT_DUR_SEC=WAIT_DUR_SEC)
        result = [f.result(internal_storage=self.internal_storage) for f in fs_dones if not f.futures]
        if self.futures:
            self.executor_futures.append(self.futures)
            self.futures = []
        self._state = ExecutorState.success
        msg = "ExecutorID {} Finished getting results".format(self.executor_id)
        logger.debug(msg)
        if result and len(result) == 1:
            return result[0]
        return result

    def create_timeline_plots(self, dst_dir, dst_file_name, futures=None):
        """
        Creates timeline and histogram of the current execution in dst_dir.

        :param futures: list of futures.
        :param dst_dir: destination folder to save .png plots.
        :param dst_file_name: name of the file.
        """
        if futures is None and (self._state == ExecutorState.new or self._state == ExecutorState.running):
            raise Exception('You must run call_async(), map() or map_reduce()'
                            ' followed by monitor() or get_results()'
                            ' before calling create_timeline_plots() method')

        if not futures:
            if self.futures:
                futures = self.futures
            elif self.executor_futures:
                futures = self.executor_futures[-1]

        if type(futures) != list:
            ftrs = [futures]
        else:
            ftrs = futures

        ftrs_to_plot = [f for f in ftrs if f.ready or f.done]

        if not ftrs_to_plot:
            return

        logging.getLogger('matplotlib').setLevel(logging.WARNING)
        from pywren_ibm_cloud.plots import create_timeline, create_histogram

        msg = 'ExecutorID {} - Creating timeline plots'.format(self.executor_id)
        logger.info(msg)
        if not self.log_level:
            print(msg)

        run_statuses = [f.run_status for f in ftrs_to_plot]
        invoke_statuses = [f.invoke_status for f in ftrs_to_plot]

        create_timeline(dst_dir, dst_file_name, self.start_time, run_statuses, invoke_statuses, self.config['ibm_cos'])
        create_histogram(dst_dir, dst_file_name, self.start_time, run_statuses, self.config['ibm_cos'])

    def clean(self, local_execution=True, delete_all=False):
        """
        Deletes all the files from COS. These files include the function,
        the data serialization and the function invocation results.
        """
        storage_bucket = self.config['pywren']['storage_bucket']
        storage_prerix = self.config['pywren']['storage_prefix']
        if delete_all:
            storage_prerix = '/'.join([storage_prerix])
        else:
            storage_prerix = '/'.join([storage_prerix, self.executor_id])
        msg = "ExecutorID {} - Cleaning temporary data from {}://{}/{}".format(self.executor_id,
                                                                               self.internal_storage.storage_backend,
                                                                               storage_bucket,
                                                                               storage_prerix)
        logger.info(msg)
        if not self.log_level:
            print(msg)

        if local_execution:
            # 1st case: Not background. The main code waits until the cleaner finishes its execution.
            # It is not ideal for performance tests, since it can take long time to complete.
            # clean_os_bucket(storage_bucket, storage_prerix, self.internal_storage)

            # 2nd case: Execute in Background as a subprocess. The main program does not wait for its completion.
            storage_config = json.dumps(self.internal_storage.get_storage_config())
            storage_config = storage_config.replace('"', '\\"')

            cmdstr = ("{} -c 'from pywren_ibm_cloud.storage.cleaner import clean_bucket; \
                              clean_bucket(\"{}\", \"{}\", \"{}\")'".format(sys.executable,
                                                                            storage_bucket,
                                                                            storage_prerix,
                                                                            storage_config))
            os.popen(cmdstr)

        else:
            extra_env = {'STORE_STATUS': False,
                         'STORE_RESULT': False}
            old_stdout = sys.stdout
            sys.stdout = open(os.devnull, 'w')
            self.executor.call_async(clean_os_bucket, [storage_bucket, storage_prerix], extra_env=extra_env)
            sys.stdout = old_stdout

        self._state = ExecutorState.finished



import os
import logging
import time
from types import SimpleNamespace
from pywren_ibm_cloud import wrenconfig
from pywren_ibm_cloud.version import __version__
from concurrent.futures import ThreadPoolExecutor
from pywren_ibm_cloud.compute import Compute
from pywren_ibm_cloud.future import ResponseFuture, JobState
from pywren_ibm_cloud.wrenconfig import extract_storage_config
from pywren_ibm_cloud.storage.storage_utils import create_output_key, create_status_key

logger = logging.getLogger(__name__)


class Invoker:

    def __init__(self, config, executor_id):
        self.log_level = os.getenv('PYWREN_LOG_LEVEL')
        self.config = config
        self.executor_id = executor_id

        compute_config = wrenconfig.extract_compute_config(self.config)
        self.internal_compute = Compute(compute_config)

    def run(self, job_description):
        job = SimpleNamespace(**job_description)
        storage_config = extract_storage_config(self.config)

        if job.remote_invocation:
            log_msg = ('ExecutorID {} - Starting {} remote invocation function: Spawning {}() '
                       '- Total: {} activations'.format(self.executor_id, job.total_data_items,
                                                        job.func_name, job.original_iterdata_len))
        else:
            log_msg = ('ExecutorID {} - Starting function invocation: {}() - Total: {} '
                       'activations'.format(self.executor_id, job.func_name, job.total_data_items))
        logger.info(log_msg)
        if not self.log_level:
            print(log_msg)

        ########################

        def invoke(executor_id, callgroup_id, call_id, func_key, invoke_metadata, data_key, data_byte_range):

            output_key = create_output_key(storage_config['storage_prefix'], executor_id, callgroup_id, call_id)
            status_key = create_status_key(storage_config['storage_prefix'], executor_id, callgroup_id, call_id)

            payload = {
                'config': self.config,
                'log_level': self.log_level,
                'func_key': func_key,
                'data_key': data_key,
                'output_key': output_key,
                'status_key': status_key,
                'job_max_runtime': job.runtime_timeout,
                'data_byte_range': data_byte_range,
                'executor_id': executor_id,
                'callgroup_id': callgroup_id,
                'call_id': call_id,
                'pywren_version': __version__}

            if job.extra_env is not None:
                logger.debug("Extra environment vars {}".format(job.extra_env))
                payload['extra_env'] = job.extra_env

            if job.extra_meta is not None:
                # sanity
                for k, v in job.extra_meta.items():
                    if k in payload:
                        raise ValueError("Key {} already in dict".format(k))
                    payload[k] = v

            # overwrite explicit args, mostly used for testing via injection
            if job.overwrite_invoke_args is not None:
                payload.update(job.overwrite_invoke_args)

            host_submit_time = time.time()
            payload['host_submit_time'] = host_submit_time
            # do the invocation
            activation_id = self.internal_compute.invoke(job.runtime_name, job.runtime_memory, payload)

            if not activation_id:
                raise Exception("ExecutorID {} - Activation {} failed, therefore job is failed".format(executor_id, call_id))

            invoke_metadata['activation_id'] = activation_id
            invoke_metadata['invoke_time'] = time.time() - host_submit_time

            invoke_metadata.update(payload)
            del invoke_metadata['config']

            fut = ResponseFuture(call_id, callgroup_id, executor_id, activation_id, storage_config, invoke_metadata)
            fut._set_state(JobState.invoked)

            return fut

        ########################

        call_futures = []
        with ThreadPoolExecutor(max_workers=job.invoke_pool_threads) as executor:
            for i in range(job.total_data_items):
                call_id = "{:05d}".format(i)
                data_byte_range = job.data_ranges[i]
                future = executor.submit(invoke, self.executor_id,
                                         job.callgroup_id, call_id, job.func_key,
                                         job.host_job_meta.copy(),
                                         job.data_key, data_byte_range)
                call_futures.append(future)

        res = [ft.result() for ft in call_futures]

        return res



import os
import json
import logging
from ..version import __version__
from .backends.ibm_cos import IbmCosStorageBackend
from .backends.swift import SwiftStorageBackend
from .exceptions import StorageNoSuchKeyError
from .storage_utils import create_status_key, create_output_key, status_key_suffix


LOCAL_HOME_DIR = os.path.expanduser('~')
logger = logging.getLogger(__name__)


class InternalStorage:
    """
    An InternalStorage object is used by executors and other components to access underlying storage backend
    without exposing the the implementation details.
    """

    def __init__(self, storage_config):

        self.storage_config = storage_config
        self.storage_backend = self.storage_config['storage_backend']
        self.storage_bucket = self.storage_config['storage_bucket']
        self.prefix = self.storage_config['storage_prefix']

        if self.storage_backend == 'ibm_cos':
            self.storage_handler = IbmCosStorageBackend(self.storage_config['ibm_cos'])
        elif self.storage_backend == 'swift':
            self.storage_handler = SwiftStorageBackend(self.storage_config['swift'])
        else:
            raise NotImplementedError(("Using {} as internal storage backend is" +
                                       "not supported yet").format(self.storage_backend))

    def get_storage_config(self):
        """
        Retrieves the configuration of this storage handler.
        :return: storage configuration
        """
        return self.storage_config

    def put_data(self, key, data):
        """
        Put data object into storage.
        :param key: data key
        :param data: data content
        :return: None
        """
        return self.storage_handler.put_object(self.storage_bucket, key, data)

    def put_func(self, key, func):
        """
        Put serialized function into storage.
        :param key: function key
        :param func: serialized function
        :return: None
        """
        return self.storage_handler.put_object(self.storage_bucket, key, func)

    def get_data(self, key, stream=False, extra_get_args={}):
        """
        Get data object from storage.
        :param key: data key
        :return: data content
        """
        return self.storage_handler.get_object(self.storage_bucket, key, stream, extra_get_args)

    def get_func(self, key):
        """
        Get serialized function from storage.
        :param key: function key
        :return: serialized function
        """
        return self.storage_handler.get_object(self.storage_bucket, key)

    def get_callset_status(self, executor_id):
        """
        Get the status of a callset.
        :param executor_id: executor's ID
        :return: A list of call IDs that have updated status.
        """
        # TODO: a better API for this is to return status for all calls in the callset. We'll fix
        #  this in scheduler refactoring.
        callset_prefix = '/'.join([self.prefix, executor_id])
        keys = self.storage_handler.list_keys_with_prefix(self.storage_bucket, callset_prefix)
        suffix = status_key_suffix
        status_keys = [k for k in keys if suffix in k]
        call_ids = [tuple(k[len(callset_prefix)+1:].split("/")[:2]) for k in status_keys]
        return call_ids

    def get_call_status(self, executor_id, callgroup_id, call_id):
        """
        Get status of a call.
        :param executor_id: executor ID of the call
        :param call_id: call ID of the call
        :return: A dictionary containing call's status, or None if no updated status
        """
        status_key = create_status_key(self.prefix, executor_id, callgroup_id, call_id)
        try:
            data = self.storage_handler.get_object(self.storage_bucket, status_key)
            return json.loads(data.decode('ascii'))
        except StorageNoSuchKeyError:
            return None

    def get_call_output(self, executor_id, callgroup_id, call_id):
        """
        Get the output of a call.
        :param executor_id: executor ID of the call
        :param call_id: call ID of the call
        :return: Output of the call.
        """
        output_key = create_output_key(self.prefix, executor_id, callgroup_id, call_id)
        try:
            return self.storage_handler.get_object(self.storage_bucket, output_key)
        except StorageNoSuchKeyError:
            return None

    def get_runtime_info(self, key):
        """
        Get the metadata given a runtime name.
        :param runtime: name of the runtime
        :return: runtime metadata
        """
        path = ['runtimes', __version__,  key+".meta.json"]
        filename_local_path = os.path.join(LOCAL_HOME_DIR, '.pywren', *path)

        if os.path.exists(filename_local_path):
            logger.debug("Runtime metadata found in local cache")
            with open(filename_local_path, "r") as f:
                runtime_meta = json.loads(f.read())
            return runtime_meta
        else:
            logger.debug("Runtime metadata not found in local cache. Retrieving it from storage")
            try:
                obj_key = '/'.join(path).replace('\\', '/')
                json_str = self.storage_handler.get_object(self.storage_bucket, obj_key)
                runtime_meta = json.loads(json_str.decode("ascii"))
                # Save runtime meta to cache
                if not os.path.exists(os.path.dirname(filename_local_path)):
                    os.makedirs(os.path.dirname(filename_local_path))

                with open(filename_local_path, "w") as f:
                    f.write(json.dumps(runtime_meta))

                return runtime_meta
            except StorageNoSuchKeyError:
                raise Exception('The runtime {} is not installed.'.format(obj_key))

    def put_runtime_info(self, key, runtime_meta):
        """
        Puit the metadata given a runtime config.
        :param runtime: name of the runtime
        :param runtime_meta metadata
        """
        path = ['runtimes', __version__,  key+".meta.json"]
        obj_key = '/'.join(path).replace('\\', '/')
        # logger.debug("Uploading Runtime metadata to: {}/{}".format(self.storage_bucket, obj_key))
        self.storage_handler.put_object(self.storage_bucket, obj_key, json.dumps(runtime_meta))

        filename_local_path = os.path.join(LOCAL_HOME_DIR, '.pywren', *path)
        # logger.debug("Saving runtime metadata in local cache: {}".format(filename_local_path))

        if not os.path.exists(os.path.dirname(filename_local_path)):
            os.makedirs(os.path.dirname(filename_local_path))

        with open(filename_local_path, "w") as f:
            f.write(json.dumps(runtime_meta))

    def delete_runtime_info(self, key):
        """
        Puit the metadata given a runtime config.
        :param runtime: name of the runtime
        :param runtime_meta metadata
        """
        path = ['runtimes', __version__,  key+".meta.json"]
        obj_key = '/'.join(path).replace('\\', '/')
        filename_local_path = os.path.join(LOCAL_HOME_DIR, '.pywren', *path)
        if os.path.exists(filename_local_path):
            os.remove(filename_local_path)
        self.storage_handler.delete_object(self.storage_bucket, obj_key)

    def list_tmp_data(self, prefix):
        """
        List the temporal data used by PyWren.
        :param bucket: bucket key
        :param prefix: prefix to search for
        :return: list of objects
        """
        return self.storage_handler.list_keys_with_prefix(self.storage_bucket, prefix)

    def delete_temporal_data(self, key_list):
        """
        Delete temporal data from PyWren.
        :param bucket: bucket name
        :param key: data key
        """
        return self.storage_handler.delete_objects(self.storage_bucket, key_list)



import os
import logging
from pywren_ibm_cloud.version import __version__
from pywren_ibm_cloud.utils import is_cf_cluster
from pywren_ibm_cloud.libs.ibm.cloudfunctions_client import CloudFunctionsClient

logger = logging.getLogger(__name__)


class IbmCfComputeBackend:
    """
    A wrap-up around IBM Cloud Functions APIs.
    """

    def __init__(self, ibm_cf_config):
        self.log_level = os.getenv('PYWREN_LOG_LEVEL')
        self.name = 'ibm_cf'
        self.ibm_cf_config = ibm_cf_config
        self.package = 'pywren_v'+__version__
        self.cf_client = CloudFunctionsClient(self.ibm_cf_config)
        self.is_cf_cluster = is_cf_cluster()

        self.region = ibm_cf_config['endpoint'].split('//')[1].split('.')[0]
        self.namespace = ibm_cf_config['namespace']
        log_msg = 'PyWren v{} init for IBM Cloud Functions - Namespace: {} - Region: {}'.format(__version__, self.namespace, self.region)
        logger.info(log_msg)
        if not self.log_level:
            print(log_msg)

    def _format_action_name(self, runtime_name, runtime_memory):
        runtime_name = runtime_name.replace('/', '_').replace(':', '_')
        return '{}_{}MB'.format(runtime_name, runtime_memory)

    def _unformat_action_name(self, action_name):
        runtime_name, memory = action_name.rsplit('_', 1)
        image_name = runtime_name.replace('_', '/', 1)
        image_name = image_name.replace('_', ':', -1)
        return image_name, int(memory.replace('MB', ''))

    def create_runtime(self, docker_image_name, memory, code=None, is_binary=True, timeout=300000):
        self.cf_client.create_package(self.package)
        action_name = self._format_action_name(docker_image_name, memory)
        self.cf_client.create_action(self.package, action_name, docker_image_name, code=code,
                                     memory=memory, is_binary=is_binary, timeout=timeout)
        return action_name

    def delete_runtime(self, docker_image_name, memory):
        action_name = self._format_action_name(docker_image_name, memory)
        self.cf_client.delete_action(self.package, action_name)

    def delete_all_runtimes(self):
        packages = self.cf_client.list_packages()
        for pkg in packages:
            if 'pywren_v' in pkg['name']:
                actions = self.cf_client.list_actions(pkg['name'])
                while actions:
                    for action in actions:
                        self.cf_client.delete_action(pkg['name'], action['name'])
                    actions = self.cf_client.list_actions(pkg['name'])
                self.cf_client.delete_package(pkg['name'])

    def list_runtimes(self, docker_image_name='all'):
        """
        List all the runtimes deployed in the IBM CF service
        return: list of tuples [docker_image_name, memory]
        """
        runtimes = []
        actions = self.cf_client.list_actions(self.package)

        for action in actions:
            action_image_name, memory = self._unformat_action_name(action['name'])
            if docker_image_name == action_image_name or docker_image_name == 'all':
                runtimes.append([action_image_name, memory])
        return runtimes

    def invoke(self, runtime_name, runtime_memory, payload):
        """
        Invoke -- return information about this invocation
        """
        action_name = self._format_action_name(runtime_name, runtime_memory)
        return self.cf_client.invoke(self.package, action_name, payload, self.is_cf_cluster)

    def invoke_with_result(self, runtime_name, runtime_memory, payload={}):
        """
        Invoke waiting for a result -- return information about this invocation
        """
        action_name = self._format_action_name(runtime_name, runtime_memory)
        return self.cf_client.invoke_with_result(self.package, action_name, payload)

    def get_runtime_key(self, runtime_name, runtime_memory):
        """
        Method that creates and returns the runtime key.
        Runtime keys are used to uniquely identify runtimes within the storage,
        in order to know which runtimes are installed and which not.
        """
        action_name = self._format_action_name(runtime_name, runtime_memory)
        runtime_key = os.path.join(self.name, self.region, self.namespace, action_name)

        return runtime_key



#
# (C) Copyright IBM Corp. 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import requests
from urllib.parse import urlencode


logger = logging.getLogger(__name__)


class IBMIAMClient:

    def __init__(self, iam_config, cf_endpoint, cf_namespace):
        self.iam_api_key = iam_config.get('api_key', None)
        self.iam_auth_endpoint = iam_config['ibm_auth_endpoint']
        self.cf_endpoint = cf_endpoint
        self.cf_namespace = cf_namespace

    def get_iam_token(self):
        data = urlencode({'grant_type': 'urn:ibm:params:oauth:grant-type:apikey', 'apikey': self.iam_api_key})
        headers = {
            'content-type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json'
        }
        res = requests.post(self.iam_auth_endpoint, data=data, headers=headers)

        if res.status_code != 200:
            raise RuntimeError("Error: http code {} while retrieving IAM token for API key.".format(res.status_code))

        bearer_response = res.json()
        bearer_token = bearer_response['access_token']
        logger.debug(bearer_token)

        return bearer_token

    def get_function_namespace_id(self, iam_token):
        logger.debug("Getting name space id for {}".format(self.cf_namespace))
        headers = {
            'content-type': 'application/json',
            'Accept': 'application/json',
            'Authorization': iam_token
        }
        url = '/'.join([self.cf_endpoint, 'api', 'v1', 'namespaces'])
        res = requests.get(url, headers=headers)
        if res.status_code != 200:
            raise RuntimeError("Error: http code {} while listing namespaces.".format(res.status_code))
        namespaces = res.json()

        for current_namespace in namespaces['namespaces']:
            if 'name' in current_namespace and current_namespace['name'] == self.cf_namespace:
                logger.debug("Found name space id {} for {}".format(current_namespace['id'], self.cf_namespace))
                return current_namespace['id']

        raise RuntimeError("Error: No CF namespace \"{}\" found.".format(self.cf_namespace))



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import time
import pika
import queue
import random
import logging
import threading
from multiprocessing.pool import ThreadPool
from .future import JobState

logger = logging.getLogger(__name__)

ALL_COMPLETED = 1
ANY_COMPLETED = 2
ALWAYS = 3


def wait(fs, executor_id, internal_storage, download_results=False,
         throw_except=True, rabbit_amqp_url=None, pbar=None,
         return_when=ALL_COMPLETED, THREADPOOL_SIZE=128, WAIT_DUR_SEC=1):
    """
    Wait for the Future instances `fs` to complete. Returns a 2-tuple of
    lists. The first list contains the futures that completed
    (finished or cancelled) before the wait completed. The second
    contains uncompleted futures.

    :param fs: A list of futures.
    :param executor_id: executor's ID.
    :param internal_storage: Storage handler to poll cloud storage.
    :param download_results: Download the results: Ture, False.
    :param rabbit_amqp_url: amqp url for accessing rabbitmq.
    :param pbar: Progress bar.
    :param return_when: One of `ALL_COMPLETED`, `ANY_COMPLETED`, `ALWAYS`
    :param THREADPOOL_SIZE: Number of threads to use. Default 64
    :param WAIT_DUR_SEC: Time interval between each check.
    :return: `(fs_dones, fs_notdones)`
        where `fs_dones` is a list of futures that have completed
        and `fs_notdones` is a list of futures that have not completed.
    :rtype: 2-tuple of lists
    """
    # FIXME:  this will eventually provide an optimization for checking if a large
    # number of futures have completed without too much network traffic
    # by exploiting the callset

    N = len(fs)
    # These are performance-related settings that we may eventually
    # want to expose to end users:
    MAX_DIRECT_QUERY_N = 64
    RETURN_EARLY_N = 32
    RANDOM_QUERY = False

    if return_when == ALL_COMPLETED:

        if rabbit_amqp_url and not download_results:
            callgroup_id = fs[0].callgroup_id
            return _wait_rabbitmq(fs, executor_id, callgroup_id, rabbit_amqp_url, pbar, N)

        result_count = 0

        while result_count < N:
            fs_dones, fs_notdones = _wait_storage(fs, executor_id,
                                                  internal_storage,
                                                  download_results,
                                                  throw_except,
                                                  RETURN_EARLY_N,
                                                  MAX_DIRECT_QUERY_N,
                                                  random_query=RANDOM_QUERY,
                                                  THREADPOOL_SIZE=THREADPOOL_SIZE,
                                                  pbar=pbar)
            N = len(fs)
            if pbar and pbar.total != N:
                pbar.total = N
                pbar.refresh()

            result_count = len(fs_dones)
            if result_count == N:
                return fs_dones, fs_notdones
            else:
                sleep = WAIT_DUR_SEC
                if fs_dones:
                    sleep = max(float(round(WAIT_DUR_SEC-((len(fs_dones)/N)*WAIT_DUR_SEC), 3)), 0)
                #print("Sleep:", sleep)
                time.sleep(sleep)
                #print('---')

    elif return_when == ANY_COMPLETED:
        while True:
            fs_dones, fs_notdones = _wait_storage(fs, executor_id,
                                                  internal_storage,
                                                  download_results,
                                                  throw_except,
                                                  RETURN_EARLY_N,
                                                  MAX_DIRECT_QUERY_N,
                                                  random_query=RANDOM_QUERY,
                                                  THREADPOOL_SIZE=THREADPOOL_SIZE)

            if len(fs_dones) != 0:
                return fs_dones, fs_notdones
            else:
                time.sleep(WAIT_DUR_SEC)

    elif return_when == ALWAYS:
        return _wait_storage(fs, executor_id,
                             internal_storage,
                             download_results,
                             throw_except,
                             RETURN_EARLY_N,
                             MAX_DIRECT_QUERY_N,
                             random_query=RANDOM_QUERY,
                             THREADPOOL_SIZE=THREADPOOL_SIZE)
    else:
        raise ValueError()


class rabbitmq_checker_worker(threading.Thread):

    def callback(self, ch, method, properties, body):
        self.q.put(body.decode("utf-8"))

    def __init__(self, executor_id, rabbit_amqp_url, q):
        threading.Thread.__init__(self)
        self.executor_id = executor_id
        self.q = q
        params = pika.URLParameters(rabbit_amqp_url)
        self.connection = pika.BlockingConnection(params)
        self.channel = self.connection.channel()  # start a channel
        self.channel.queue_declare(queue=self.executor_id, auto_delete=True)
        self.channel.basic_consume(self.callback, queue=self.executor_id, no_ack=True)

    def run(self):
        msg = 'ExecutorID {} - Starting consumer from rabbitmq queue'.format(self.executor_id)
        logger.debug(msg)
        self.channel.start_consuming()

    def __del__(self):
        self.channel.queue_delete(queue=self.executor_id)
        self.channel.stop_consuming()


def _wait_rabbitmq(fs, executor_id, callgroup_id, rabbit_amqp_url, pbar, total):
    q = queue.Queue()
    td = rabbitmq_checker_worker(executor_id, rabbit_amqp_url, q)
    td.setDaemon(True)
    td.start()

    done_call_status = {}
    done_call_status[callgroup_id] = {}
    done_call_ids = {}
    done_call_ids[callgroup_id] = {'total': total, 'call_ids': []}

    def call_ids_to_futures():
        fs_dones = []
        fs_notdones = []
        for f in fs:
            if f.callgroup_id in done_call_status and f.call_id in done_call_status[f.callgroup_id]:
                f.run_status = done_call_status[f.callgroup_id][f.call_id]
                f.invoke_status['status_done_timestamp'] = f.run_status['status_done_timestamp']
                del f.run_status['status_done_timestamp']
                f._set_state(JobState.ready)
                fs_dones.append(f)
            else:
                fs_notdones.append(f)
        return fs_dones, fs_notdones

    def reception_finished():
        for cg_id in done_call_ids:
            total = done_call_ids[cg_id]['total']
            recived_call_ids = len(done_call_ids[cg_id]['call_ids'])

            if total is None or total > recived_call_ids:
                return False

        return True

    while not reception_finished():
        try:
            call_status = json.loads(q.get())
            call_status['status_done_timestamp'] = time.time()
        except KeyboardInterrupt:
            call_ids_to_futures()
            raise KeyboardInterrupt

        rcv_callgroup_id = call_status['callgroup_id']
        rcv_call_id = call_status['call_id']

        if rcv_callgroup_id not in done_call_ids:
            done_call_ids[rcv_callgroup_id] = {'total': None, 'call_ids': []}
        if rcv_callgroup_id not in done_call_status:
            done_call_status[rcv_callgroup_id] = {}

        done_call_ids[rcv_callgroup_id]['call_ids'].append(rcv_call_id)
        done_call_status[rcv_callgroup_id][rcv_call_id] = call_status

        if pbar:
            pbar.update(1)
            pbar.refresh()

        if 'new_futures' in call_status:
            new_futures = call_status['new_futures'].split('/')
            if int(new_futures[1]) != 0:
                # We received new futures to track
                callgroup_id_new_futures = new_futures[0]
                total_new_futures = int(new_futures[1])
                if callgroup_id_new_futures not in done_call_ids:
                    done_call_ids[callgroup_id_new_futures] = {'total': total_new_futures, 'call_ids': []}
                else:
                    done_call_ids[callgroup_id_new_futures]['total'] = total_new_futures

                if pbar:
                    pbar.total = pbar.total + total_new_futures
                    pbar.refresh()
    if pbar:
        pbar.close()

    return call_ids_to_futures()


def _wait_storage(fs, executor_id, internal_storage, download_results,
                  throw_except, return_early_n, max_direct_query_n,
                  random_query=False, THREADPOOL_SIZE=128, pbar=None):
    """
    internal function that performs the majority of the WAIT task
    work.

    For the list of futures fn, we will check at a minimum `max_direct_query_n`
    futures at least once. Internally we :
    1. use list() to quickly get a list of which ones are done (but
    list can be behind due to eventual consistency issues)
    2. then individually call get_status on at most `max_direct_query_n` returning
       early if we have found at least `return_early_n`

    This can mitigate the stragglers.

    random_query decides whether we get the fs in the order they are presented
    or in a random order.
    """
    # get all the futures that are not yet done
    if download_results:
        not_done_futures = [f for f in fs if not f.done]
    else:
        not_done_futures = [f for f in fs if not f.ready]

    if len(not_done_futures) == 0:
        return fs, []

    # note this returns everything done, so we have to figure out
    # the intersection of those that are done
    #t0 = time.time()
    callids_done_in_callset = set(internal_storage.get_callset_status(executor_id))
    #print('Time getting list: ', time.time()-t0, len(callids_done_in_callset))
    # print('CALLSET:', callids_done_in_callset, len(callids_done_in_callset))

    not_done_call_ids = set([(f.callgroup_id, f.call_id) for f in not_done_futures])
    # print('NO TDONE:' ,not_done_call_ids, len(not_done_call_ids))

    done_call_ids = not_done_call_ids.intersection(callids_done_in_callset)
    not_done_call_ids = not_done_call_ids - done_call_ids
    still_not_done_futures = [f for f in not_done_futures if ((f.callgroup_id, f.call_id) in not_done_call_ids)]

    def fetch_future_status(f):
        return internal_storage.get_call_status(f.executor_id, f.callgroup_id, f.call_id)

    pool = ThreadPool(THREADPOOL_SIZE)

    # now try up to max_direct_query_n direct status queries, quitting once
    # we have return_n done.
    query_count = 0
    max_queries = min(max_direct_query_n, len(still_not_done_futures))

    if random_query:
        random.shuffle(still_not_done_futures)

    while query_count < max_queries:
        if len(done_call_ids) >= return_early_n:
            break
        num_to_query_at_once = THREADPOOL_SIZE
        fs_to_query = still_not_done_futures[query_count:query_count + num_to_query_at_once]

        fs_statuses = pool.map(fetch_future_status, fs_to_query)

        callids_found = [(fs_to_query[i].callgroup_id, fs_to_query[i].call_id) for i in range(len(fs_to_query))
                         if fs_statuses[i] is not None]

        # print('FOUND:', callids_found, len(callids_found))

        done_call_ids = done_call_ids.union(set(callids_found))
        query_count += len(fs_to_query)

    # now we walk through all the original queries and get
    # the ones that are actually done.
    fs_dones = []
    fs_notdones = []
    f_to_wait_on = []
    for f in fs:
        if f.ready or f.done:
            # done, don't need to do anything
            fs_dones.append(f)
        else:
            if (f.callgroup_id, f.call_id) in done_call_ids:
                f_to_wait_on.append(f)
                fs_dones.append(f)
            else:
                fs_notdones.append(f)

#     if still_not_done_futures and len(still_not_done_futures) < max(1, int(len(fs)*0.015)):
#         f_to_wait_on.extend(still_not_done_futures)
#         fs_dones.extend(still_not_done_futures)

    def get_result(f):
        f.result(throw_except=throw_except, internal_storage=internal_storage)
        #if pbar and f.done:
        #    pbar.update(1)

    def get_status(f):
        f.status(throw_except=throw_except, internal_storage=internal_storage)
        #if pbar and f.ready:
        #    pbar.update(1)

#     with ThreadPoolExecutor(max_workers=THREADPOOL_SIZE) as executor:
#         if download_results:
#             executor.map(get_result, f_to_wait_on)
#         else:
#             executor.map(get_status, f_to_wait_on)

    if download_results:
        pool.map(get_result, f_to_wait_on)
    else:
        pool.map(get_status, f_to_wait_on)

    if pbar:
        for f in f_to_wait_on:
            if f.ready or f.done:
                pbar.update(1)
        pbar.refresh()
    pool.close()
    pool.join()

    # Check for new futures
    new_futures = [f.result() for f in f_to_wait_on if f.futures]
    for futures in new_futures:
        fs.extend(futures)

    return fs_dones, fs_notdones



#
# (C) Copyright IBM Corp. 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import os
import json
import struct
import logging
import requests
import inspect
from pywren_ibm_cloud import utils
from multiprocessing.pool import ThreadPool
from pywren_ibm_cloud.storage.backends.ibm_cos import IbmCosStorageBackend

logger = logging.getLogger(__name__)

CHUNK_THRESHOLD = 128*1024  # 128KB


def partition_processor(map_function):
    """
    Method that returns the function to process objects in the Cloud.
    It creates a ready-to-use data_stream parameter
    """
    def object_processing_wrapper(map_func_args, data_byte_range, chunk_size, ibm_cos):
        extra_get_args = {}
        if data_byte_range is not None:
            range_str = 'bytes={}-{}'.format(*data_byte_range)
            extra_get_args['Range'] = range_str

        if 'url' in map_func_args:
            # it is a public url
            resp = requests.get(map_func_args['url'], headers=extra_get_args, stream=True)
            map_func_args['data_stream'] = resp.raw

        elif 'key' in map_func_args:
            # it is a COS key
            if 'bucket' not in map_func_args or ('bucket' in map_func_args and not map_func_args['bucket']):
                bucket, key = map_func_args['key'].split('/', 1)
            else:
                bucket = map_func_args['bucket']
                key = map_func_args['key']

            config = json.loads(os.environ.get('PYWREN_CONFIG'))
            storage = IbmCosStorageBackend(config['ibm_cos'])
            logger.info('Getting dataset from cos://{}/{}'.format(bucket, key))
            sb = storage.get_object(bucket, key, stream=True, extra_get_args=extra_get_args)
            if data_byte_range is not None:
                logger.info('Chunk range: {}'.format(extra_get_args['Range']))
                map_func_args['data_stream'] = WrappedStreamingBodyPartition(sb, chunk_size, data_byte_range)
            else:
                map_func_args['data_stream'] = sb

        func_sig = inspect.signature(map_function)
        if 'ibm_cos' in func_sig.parameters:
            map_func_args['ibm_cos'] = ibm_cos

        return map_function(**map_func_args)

    return object_processing_wrapper


def create_partitions(config, arg_data, chunk_size):
    """
    Method that returns the function that will create the partitions of the objects in the Cloud
    """
    logger.debug('Starting partitioner')

    # We suppose here that the data is always in IBM COS.  TODO: Make it Generic.
    storage = IbmCosStorageBackend(config['ibm_cos'])

    map_func_keys = arg_data[0].keys()
    parts_per_object = None

    if 'bucket' in map_func_keys and 'key' not in map_func_keys:
        partitions, parts_per_object = split_objects_from_bucket(arg_data, chunk_size, storage)
        if not partitions:
            raise Exception('No objects available within bucket: {}'.format(arg_data[0]['bucket']))

    elif 'key' in map_func_keys:
        partitions, parts_per_object = split_object_from_key(arg_data, chunk_size, storage)

    elif 'url' in map_func_keys:
        partitions, parts_per_object = split_object_from_url(arg_data, chunk_size)

    else:
        raise ValueError('You did not provide any bucket or object key/url')

    return partitions, parts_per_object


def split_objects_from_bucket(map_func_args_list, chunk_size, storage):
    """
    Create partitions from bucket/s
    """
    logger.info('Creating dataset chunks from bucket/s ...')
    partitions = []
    parts_per_object = []

    for entry in map_func_args_list:
        # Each entry is a bucket
        if chunk_size:
            logger.info('Creating chunks from objects within: {}'.format(entry['bucket']))
        else:
            logger.info('Discovering objects within: {}'.format(entry['bucket']))
        bucket_name, prefix = utils.split_path(entry['bucket'])
        objects = storage.list_objects(bucket_name, prefix)

        def _split(obj):
            key = obj['Key']
            obj_size = obj['Size']
            total_partitions = 0
            #logger.info("Extracted key {} size {}".format(key, obj_size))

            # full_key = '{}/{}'.format(bucket_name, key)
            size = 0
            if chunk_size is not None and obj_size > chunk_size:
                while size < obj_size:
                    brange = (size, size+chunk_size+CHUNK_THRESHOLD)
                    size += chunk_size
                    partition = {}
                    partition['map_func_args'] = entry.copy()
                    partition['map_func_args']['key'] = key
                    partition['map_func_args']['bucket'] = bucket_name
                    partition['data_byte_range'] = brange
                    partition['chunk_size'] = chunk_size
                    partitions.append(partition)
                    total_partitions = total_partitions + 1
            else:
                partition = {}
                partition['map_func_args'] = entry.copy()
                partition['map_func_args']['key'] = key
                partition['map_func_args']['bucket'] = bucket_name
                partition['data_byte_range'] = None
                partition['chunk_size'] = obj_size
                partitions.append(partition)
                total_partitions = total_partitions + 1

            parts_per_object.append(total_partitions)

        pool = ThreadPool(128)
        pool.map(_split, objects)
        pool.close()
        pool.join()

    return partitions, parts_per_object


def split_object_from_key(map_func_args_list, chunk_size, storage):
    """
    Create partitions from a list of COS objects keys
    """
    if chunk_size:
        logger.info('Creating chunks from object keys...')
    partitions = []
    parts_per_object = []

    def _split(entry):
        total_partitions = 0
        object_key = entry['key']
        bucket, object_name = object_key.split('/', 1)
        metadata = storage.head_object(bucket, object_name)
        obj_size = int(metadata['content-length'])

        if chunk_size is not None and obj_size > chunk_size:
            size = 0
            while size < obj_size:
                brange = (size, size+chunk_size+CHUNK_THRESHOLD)
                size += chunk_size
                partition = {}
                partition['map_func_args'] = entry
                partition['data_byte_range'] = brange
                partition['chunk_size'] = chunk_size
                partitions.append(partition)
                total_partitions = total_partitions + 1
        else:
            partition = {}
            partition['map_func_args'] = entry
            partition['data_byte_range'] = None
            partition['chunk_size'] = obj_size
            partitions.append(partition)
            total_partitions = total_partitions + 1

        parts_per_object.append(total_partitions)

    pool = ThreadPool(128)
    pool.map(_split, map_func_args_list)
    pool.close()
    pool.join()

    return partitions, parts_per_object


def split_object_from_url(map_func_args_list, chunk_size):
    """
    Create partitions from a list of objects urls
    """
    if chunk_size:
        logger.info('Creating chunks from urls...')
    partitions = []
    parts_per_object = []

    def _split(entry):
        obj_size = None
        total_partitions = 0
        object_url = entry['url']
        metadata = requests.head(object_url)

        logger.info(object_url)
        #logger.debug(metadata.headers)

        if 'content-length' in metadata.headers:
            obj_size = int(metadata.headers['content-length'])

        if 'accept-ranges' in metadata.headers and chunk_size is not None \
           and obj_size is not None and obj_size > chunk_size:
            size = 0
            while size < obj_size:
                brange = (size, size+chunk_size+CHUNK_THRESHOLD)
                size += chunk_size
                partition = {}
                partition['map_func_args'] = entry
                partition['data_byte_range'] = brange
                partition['chunk_size'] = chunk_size
                partitions.append(partition)
                total_partitions = total_partitions + 1
        else:
            # Only one partition
            partition = {}
            partition['map_func_args'] = entry
            partition['data_byte_range'] = None
            partition['chunk_size'] = obj_size
            partitions.append(partition)
            total_partitions = total_partitions + 1

        parts_per_object.append(total_partitions)

    pool = ThreadPool(128)
    pool.map(_split, map_func_args_list)
    pool.close()
    pool.join()

    return partitions, parts_per_object


class WrappedStreamingBodyPartition(utils.WrappedStreamingBody):

    def __init__(self, sb, size, byterange):
        super().__init__(sb, size)
        # Range of the chunk
        self.range = byterange
        # The first chunk does not contain plusbyte
        self.plusbytes = 0 if not self.range or self.range[0] == 0 else 1
        # To store the first byte of this chunk, which actually is the last byte of previous chunk
        self.first_byte = None
        # Flag that indicates the end of the file
        self.eof = False

    def read(self, n=None):
        if self.eof:
            raise EOFError()
        # Data always contain one byte from the previous chunk,
        # so l'ets check if it is a \n or not
        self.first_byte = self.sb.read(self.plusbytes)
        retval = self.sb.read(n)

        if retval == "":
            raise EOFError()

        self.pos += len(retval)

        first_row_start_pos = 0
        if self.first_byte != b'\n' and self.plusbytes != 0:
            logger.debug('Discarding first partial row')
            # Previous byte is not \n
            # This means that we have to discard first row because it is cut
            first_row_start_pos = retval.find(b'\n')+1

        last_row_end_pos = self.pos
        # Find end of the line in threshold
        if self.pos > self.size:
            buf = io.BytesIO(retval[self.size:])
            buf.readline()
            last_row_end_pos = self.size+buf.tell()
            self.eof = True

        return retval[first_row_start_pos:last_row_end_pos]

    def readline(self):
        if self.eof:
            raise EOFError()

        if not self.first_byte and self.plusbytes != 0:
            self.first_byte = self.sb.read(self.plusbytes)
            if self.first_byte != b'\n':
                logger.debug('Discarding first partial row')
                self.sb._raw_stream.readline()
        try:
            retval = self.sb._raw_stream.readline()
        except struct.error:
            raise EOFError()
        self.pos += len(retval)

        if self.pos >= self.size:
            self.eof = True

        return retval



#
# (C) Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import json
from . import InternalStorage

logger = logging.getLogger(__name__)


def clean_bucket(bucket, prefix, storage_config):
    """
    Wrapper of clean_os_bucket(). Use this method only when storage_config is
    in JSON format. In any other case, call directly clean_os_bucket() method.
    """
    internal_storage = InternalStorage(json.loads(storage_config))
    # sys.stdout = open(os.devnull, 'w')
    clean_os_bucket(bucket, prefix, internal_storage)
    # sys.stdout = sys.__stdout__


def clean_os_bucket(bucket, prefix, internal_storage):
    """
    Deletes all the files from COS. These files include the function,
    the data serialization and the function invocation results.
    """
    msg = "Going to delete all objects from bucket '{}' and prefix '{}'".format(bucket, prefix)
    logger.debug(msg)
    total_objects = 0
    objects_to_delete = internal_storage.list_tmp_data(prefix)

    while objects_to_delete:
        logger.debug('{} objects found'.format(len(objects_to_delete)))
        total_objects = total_objects + len(objects_to_delete)
        internal_storage.delete_temporal_data(objects_to_delete)
        objects_to_delete = internal_storage.list_tmp_data(prefix)
    logger.info('Finished deleting objects, total found: {}'.format(total_objects))



#
# Copyright 2018 PyWren Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import os
import sys
import uuid
import inspect
import subprocess
import struct
import platform


def uuid_str():
    return str(uuid.uuid4())


def create_executor_id(lenght=13):
    return uuid_str()[0:lenght]


def create_callgroup_id(lenght=8):
    return uuid_str().replace('-', '')[0:lenght]


def timeout_handler(signum, frame):
    raise TimeoutError()


def version_str(version_info):
    return "{}.{}".format(version_info[0], version_info[1])


def is_unix_system():
    curret_system = platform.system()
    return curret_system != 'Windows'


def is_cf_cluster():
    """
    Checks if the current execution is in an OpenWhisk function
    """
    if any([k.startswith('__OW_') for k in os.environ.keys()]):
        return True
    return False


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def is_object_processing_function(map_function):
    func_sig = inspect.signature(map_function)
    return {'bucket', 'key', 'url'} & set(func_sig.parameters)


def iterdata_as_list(iterdata):
    """
    Converts iteradat to a list
    """
    if type(iterdata) == str:
        return [iterdata]
    elif type(iterdata) != list:
        return list(iterdata)
    else:
        return iterdata


def convert_bools_to_string(extra_env):
    """
    Converts all booleans of a dictionary to a string
    """
    for key in extra_env:
        if type(extra_env[key]) == bool:
            extra_env[key] = str(extra_env[key])

    return extra_env


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def sdb_to_dict(item):
    attr = item['Attributes']
    return {c['Name']: c['Value'] for c in attr}


def bytes_to_b64str(byte_data):
    byte_data_64 = base64.b64encode(byte_data)
    byte_data_64_ascii = byte_data_64.decode('ascii')
    return byte_data_64_ascii


def b64str_to_bytes(str_data):
    str_ascii = str_data.encode('ascii')
    byte_data = base64.b64decode(str_ascii)
    return byte_data


def split_s3_url(s3_url):
    if s3_url[:5] != "s3://":
        raise ValueError("URL {} is not valid".format(s3_url))

    splits = s3_url[5:].split("/")
    bucket_name = splits[0]
    key = "/".join(splits[1:])
    return bucket_name, key


def split_path(path):

    if (path.startswith("/")):
        path = path[1:]
    ind = path.find("/")
    if (ind > 0):
        bucket_name = path[:ind]
        key = path[ind + 1:]
    else:
        bucket_name = path
        key = None
    return bucket_name, key


def get_current_memory_usage():
    """
    Gets the current memory usage of the runtime.
    To be used only in the action code.
    """
    #cmd = "ps -eo pss | tail -n +2 | awk '{ SUM += $1} END { print SUM/1024}'"
    #memorymbytes = subprocess.check_output(cmd, shell=True).decode("ascii").strip()

    cmd = "python pywren_ibm_cloud/libs/ps_mem/ps_mem.py --total"
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    memorybytes = int(process.stdout.read())

    return sizeof_fmt(memorybytes)


class WrappedStreamingBody:
    """
    Wrap boto3's StreamingBody object to provide enough Python fileobj functionality,
    and to discard data added by partitioner and cut lines.

    from https://gist.github.com/debedb/2e5cbeb54e43f031eaf0

    """
    def __init__(self, sb, size):
        # The StreamingBody we're wrapping
        self.sb = sb
        # Initial position
        self.pos = 0
        # Size of the object
        self.size = size

    def tell(self):
        # print("In tell()")
        return self.pos

    def read(self, n=None):
        retval = self.sb.read(n)
        if retval == "":
            raise EOFError()
        self.pos += len(retval)
        return retval

    def readline(self):
        try:
            retval = self.sb.readline()
        except struct.error:
            raise EOFError()
        self.pos += len(retval)
        return retval

    def seek(self, offset, whence=0):
        # print("Calling seek()")
        retval = self.pos
        if whence == 2:
            if offset == 0:
                retval = self.size
            else:
                raise Exception("Unsupported")
        else:
            if whence == 1:
                offset = self.pos + offset
                if offset > self.size:
                    retval = self.size
                else:
                    retval = offset
        # print("In seek(%s, %s): %s, size is %s" % (offset, whence, retval, self.size))

        self.pos = retval
        return retval

    def __str__(self):
        return "WrappedBody"

    def __getattr__(self, attr):
        # print("Calling %s"  % attr)

        if attr == 'tell':
            return self.tell
        elif attr == 'seek':
            return self.seek
        elif attr == 'read':
            return self.read
        elif attr == 'readline':
            return self.readline
        elif attr == '__str__':
            return self.__str__
        else:
            return getattr(self.sb, attr)


def verify_args(func, data, object_processing=False):
    # Verify parameters
    non_verify_args = ['ibm_cos', 'swift', 'internal_storage']
    func_sig = inspect.signature(func)

    # Check mandatory parameters in function
    if object_processing:
        non_verify_args.append('data_stream')
        err_msg = 'parameter in your map_function() is mandatory for pywren.map_reduce(map_function,...)'
        if 'bucket' in func_sig.parameters:
            non_verify_args.append('key')
            if 'key' not in func_sig.parameters:
                raise ValueError('"key" {}'.format(err_msg))
            if 'data_stream' not in func_sig.parameters:
                raise ValueError('"data_stream" {}'.format(err_msg))
        if 'key' in func_sig.parameters or 'url' in func_sig.parameters:
            if 'data_stream' not in func_sig.parameters:
                raise ValueError('"data_stream" {}'.format(err_msg))

    new_parameters = list()
    for param in func_sig.parameters:
        if func_sig.parameters[param].default is not None and param not in non_verify_args:
            new_parameters.append(func_sig.parameters[param])

    new_func_sig = func_sig.replace(parameters=new_parameters)

    new_data = list()
    for elem in data:
        if type(elem) == dict:
            if set(list(new_func_sig.parameters.keys())) <= set(elem):
                new_data.append(elem)
            else:
                raise ValueError("Check the args names in the data. "
                                 "You provided these args: {}, and "
                                 "the args must be: {}".format(list(elem.keys()),
                                                               list(new_func_sig.parameters.keys())))
        elif type(elem) in (list, tuple):
            new_elem = dict(new_func_sig.bind(*list(elem)).arguments)
            new_data.append(new_elem)
        else:
            # single value (string, integer, etc)
            new_elem = dict(new_func_sig.bind(elem).arguments)
            new_data.append(new_elem)

    return new_data
